# pylint: disable=protected-access

import os
from unittest.mock import patch
import joblib
import pytest
import numpy as np
from dodfminer.extract.polished.backend.ner import ActNER


@pytest.fixture(name='act_ner_with_model')
@patch.object(ActNER, '_load_model',
              return_value=joblib.load(""+os.path.dirname(__file__)+"/support/test_model.pkl"))
def fixture_act_ner_with_model(_):
    act = ActNER()
    act._name = ""
    act._backend = 'ner'
    return act


def test_act_ner_backend_ner(act_ner_with_model):
    assert act_ner_with_model._backend == 'ner'


def test_act_ner_preprocess(act_ner_with_model):
    # Caso 1 (Sentenca inventada)
    str1 = "hoje, eu vou. amanha, talvez. pode ser\n que sim; pode\n ser que nao"
    str2 = ["hoje", ",", "eu", "vou", ".", "amanha", ",", "talvez", ".",
            "pode", "ser", "que", "sim", ";", "pode", "ser", "que", "nao"]
    test1 = (act_ner_with_model._preprocess(str1) == str2)
    # Caso 2 (Sentenca de um DODF)
    str1 = "Art. 4o Validar os atos escolares praticados pela instituicao educacional, a contar de 1o de\njaneiro de 2012 ate " + \
        "a data de publicacao da portaria oriunda do presente parec e r. "
    str2 = [
        "Art",
        ".",
        "4o",
        "Validar",
        "os",
        "atos",
        "escolares",
        "praticados",
        "pela",
        "instituicao",
        "educacional",
        ",",
        "a",
        "contar",
        "de",
        "1o",
        "de",
        "janeiro",
        "de",
        "2012",
        "ate",
        "a",
        "data",
        "de",
        "publicacao",
        "da",
        "portaria",
        "oriunda",
        "do",
        "presente",
        "parec",
        "e",
        "r",
        "."
    ]
    test2 = (act_ner_with_model._preprocess(str1) == str2)
    assert test1 and test2


def test_act_ner_get_features(act_ner_with_model):

    list_1 = ["Art", ".", "4", "Validar", "ATOS", "."]
    features_1 = [
		{'bias': 1.0,
		 'word': 'art',
		 'is_title': True,
		 'is_upper': False,
		 'is_digit': False,
		 'num_digits': '0',
		 'has_hyphen': False,
		 'has_dot': False,
		 'has_slash': False,
		 'BOS': True,
		 '+1:word': '.',
		 '+1:title': False,
		 '+1:upper': False,
		 '+1:num_digits': '0',
		 '+2:word': '4',
		 '+2:title': False,
		 '+2:upper': False,
		 '+2:num_digits': '1',
		 '+3:word': 'validar',
		 '+3:title': True,
		 '+3:upper': False,
		 '+3:num_digits': '0',
		},
		{'bias': 1.0,
		 'word': '.',
		 'is_title': False,
		 'is_upper': False,
		 'is_digit': False,
		 'num_digits': '0',
		 'has_hyphen': False,
		 'has_dot': True,
		 'has_slash': False,
		 '-1:word': 'art',
		 '-1:title': True,
		 '-1:upper': False,
		 '-1:num_digits': '0',
		 '+1:word': '4',
		 '+1:title': False,
		 '+1:upper': False,
		 '+1:num_digits': '1',
		 '+2:word': 'validar',
		 '+2:title': True,
		 '+2:upper': False,
		 '+2:num_digits': '0',
		 '+3:word': 'atos',
		 '+3:title': False,
		 '+3:upper': True,
		 '+3:num_digits': '0',
		},
		{'bias': 1.0,
		 'word': '4',
		 'is_title': False,
		 'is_upper': False,
		 'is_digit': True,
		 'num_digits': '1',
		 'has_hyphen': False,
		 'has_dot': False,
		 'has_slash': False,
		 '-1:word': '.',
		 '-1:title': False,
		 '-1:upper': False,
		 '-1:num_digits': '0',
		 '-2:word': 'art',
		 '-2:title': True,
		 '-2:upper': False,
		 '-2:num_digits': '0',
		 '+1:word': 'validar',
		 '+1:title': True,
		 '+1:upper': False,
		 '+1:num_digits': '0',
		 '+2:word': 'atos',
		 '+2:title': False,
		 '+2:upper': True,
		 '+2:num_digits': '0',
		 '+3:word': '.',
		 '+3:title': False,
		 '+3:upper': False,
		 '+3:num_digits': '0',
		},
		{'bias': 1.0,
		 'word': 'validar',
		 'is_title': True,
		 'is_upper': False,
		 'is_digit': False,
		 'num_digits': '0',
		 'has_hyphen': False,
		 'has_dot': False,
		 'has_slash': False,
		 '-1:word': '4',
		 '-1:title': False,
		 '-1:upper': False,
		 '-1:num_digits': '1',
		 '-2:word': '.',
		 '-2:title': False,
		 '-2:upper': False,
		 '-2:num_digits': '0',
		 '-3:word': 'art',
		 '-3:title': True,
		 '-3:upper': False,
		 '-3:num_digits': '0',
		 '+1:word': 'atos',
		 '+1:title': False,
		 '+1:upper': True,
		 '+1:num_digits': '0',
		 '+2:word': '.',
		 '+2:title': False,
		 '+2:upper': False,
		 '+2:num_digits': '0',
		},
		{'bias': 1.0,
		 'word': 'atos',
		 'is_title': False,
		 'is_upper': True,
		 'is_digit': False,
		 'num_digits': '0',
		 'has_hyphen': False,
		 'has_dot': False,
		 'has_slash': False,
		 '-1:word': 'validar',
		 '-1:title': True,
		 '-1:upper': False,
		 '-1:num_digits': '0',
		 '-2:word': '4',
		 '-2:title': False,
		 '-2:upper': False,
		 '-2:num_digits': '1',
		 '-3:word': '.',
		 '-3:title': False,
		 '-3:upper': False,
		 '-3:num_digits': '0',
		 '+1:word': '.',
		 '+1:title': False,
		 '+1:upper': False,
		 '+1:num_digits': '0',
		},
		{'bias': 1.0,
		 'word': '.',
		 'is_title': False,
		 'is_upper': False,
		 'is_digit': False,
		 'num_digits': '0',
		 'has_hyphen': False,
		 'has_dot': True,
		 'has_slash': False,
		 '-1:word': 'atos',
		 '-1:title': False,
		 '-1:upper': True,
		 '-1:num_digits': '0',
		 '-2:word': 'validar',
		 '-2:title': True,
		 '-2:upper': False,
		 '-2:num_digits': '0',
		 '-3:word': '4',
		 '-3:title': False,
		 '-3:upper': False,
		 '-3:num_digits': '1',
		 'EOS': True,
		},
    ]
    assert features_1 == act_ner_with_model._get_features(list_1)


def test_act_ner_prediction(act_ner_with_model):
    sentence = "matricula 190133743 joao souza artigo 5o data 03/12/1901 cargo professor"
    # Step-by-step execution of _prediction()
    sent = act_ner_with_model._preprocess(sentence)
    feats = act_ner_with_model._get_features(sent)
    predictions = act_ner_with_model._model.predict_single(feats)
    entities_predicted_step_by_step = act_ner_with_model._predictions_dict(
        sent, predictions)
    # Direct execution of _prediction()
    entities_predicted_at_once = act_ner_with_model._prediction(sentence)

    # Evaluating _model.predict_single()
    test1 = True
    for prediction in predictions:
        if prediction not in act_ner_with_model._model.classes_:
            test1 = False
    # Check if step-by-step is equivalent to at_once execution
    test2 = entities_predicted_at_once == entities_predicted_step_by_step
    assert test1 and test2


def test_act_ner_predictions_dict(act_ner_with_model):
    prediction = ["B-matricula", "I-matricula", "B-Nome do Servidor", "I-Nome do Servidor", "B-fundamento legal do abono de permanencia",
                  "I-fundamento legal do abono de permanencia", "B-Vigencia", "I-Vigencia", "B-Cargo Efetivo", "I-Cargo Efetivo"]
    sentence = ["matricula", "190133743", "joao", "souza",
                "artigo", "5o",  "data", "03/12/1901", "cargo", "professor"]
    entidades = {
        'Tipo do Ato': '',
        'matricula': "matricula 190133743",
        'Nome do Servidor': "joao souza",
        'fundamento legal do abono de permanencia': "artigo 5o",
        'Vigencia': "data 03/12/1901",
        'Cargo Efetivo': "cargo professor",
        'Classe': np.nan,
        'padrao': np.nan,
        'Quadro pessoal permanente ou Suplementar': np.nan,
        'Matricula SIAPE': np.nan,
        'Processo GDF/SEI': np.nan,
        'orgao': np.nan
    }
    assert entidades == act_ner_with_model._predictions_dict(
        sentence, prediction)



"""Regras regex para ato de Extrato de Contrato."""

import re
import os
import joblib
import pandas as pd
from dodfminer.extract.polished.acts.base import Atos


class Contratos(Atos):
     
    def __init__(self, file, backend):
        super().__init__(file, backend)

    def _regex_flags(self):
        return re.IGNORECASE

    def _load_model(self):
        f_path = os.path.dirname(__file__)
        f_path += '/models/efetivos_ret.pkl'
        return joblib.load(f_path)

    def _act_name(self):
        return "Contrato"

    def _props_names(self):
        return ["Tipo do Ato", "Objeto"]

    def _rule_for_inst(self):
        start = r"(EXTRATO DO CONTRATO\s)"
        body = r"([\s\S]*?"
        end = r"<EOB>)"
        return start + body + end

    def _prop_rules(self):
        regra_objeto = r"(?:OBJETO:)([\s\S]*?)[.]"

        rules = {'Objeto':regra_objeto
                 }
        return rules

    def _regex_instances(self):
        results = ContractExtractorREGEX.extract_text(self._text)

        return results

class ContractExtractorREGEX:
    """Extract contract statements from a string and returns the contracts found in a list.

    Extracts contract statements from DODF dataframe through REGEX patterns.

    Note:
        This class is not constructable, it cannot generate objects.

    """
    
    @classmethod
    def extract_text(cls, txt_string):
        """Extract texts of contract statements from dataframe or file

        Args:
            txt_string: The string from where to extract the contracts.

        Returns:
            List with the contracts extracted from the string passed.
        """
        
        base_str = txt_string

        contract_pattern_1 = r"(\nxx([a-z]{0,10})\sEXTRAT([A-Z]{0,3})\sD([A-Z]{0,3})\sCONTRAT([A-Z]{0,3}))"
        contract_pattern_2 = r"(\nxx([a-z]{0,10})\sEXTRAT([A-Z]{0,3})\sCONTRAT([A-Z]{0,3}))"

        contract_pattern = [contract_pattern_1, contract_pattern_2]
        contract_pattern = "|".join(contract_pattern)
        p_ext = re.compile(contract_pattern)

        block_pattern = r"(\nxx([a-z]{0,10})\s([A-ZÀ-Ú0-9º\/x\-\.\,\(\)\|\*&\s\']{5,}) xx([a-z]{0,10}))"
        block_pattern = block_pattern + "|" + contract_pattern
        p_blk = re.compile(block_pattern)

        nl_pattern = r"\n"
        p_nl = re.compile(nl_pattern)

        extracted_texts = cls._extract_text_blocks(base_str, p_ext, p_blk, p_nl)

        # Padrões de começo e fim de página abarcam apenas os padrões observados entre 2000 e 2021
        start_page_patterns = [r"\nPÁGINA\s([0-9]{1,5})", r"\nDIÁRIO\sOFICIAL\sDO\sDISTRITO\sFEDERAL",
                               r"\nNº(.+?)2([0-9]{3})", r"\nxx([a-z]{0,10}) Diário Oficial do Distrito Federal xx([a-z]{0,10})",
                               r"\nDiário Oficial do Distrito Federal"]

        end_page_patterns = [r"Documento assinado digitalmente conforme MP nº 2.200-2 de 24/08/2001, que institui a",
                                r"Infraestrutura de Chaves Públicas Brasileira ICP-Brasil",
                                r"Este documento pode ser verificado no endereço eletrônico",
                                r"http://wwwin.gov.br/autenticidade.html",
                                r"http://www.in.gov.br/autenticidade.html",
                                r"pelo código ([0-9]{15,18})",
                                r"\nDocumento assinado digitalmente, original em https://www.dodf.df.gov.br"]

        middle_page_patterns = [r"xx([a-z]{1,10}) ", r" xx([a-z]{1,10})", r"\n-\n",
                             r"xx([a-z]{1,10})", r"\n- -\n", r"\n- - -\n",
                            r"\n[\.\,\-\—]\n", r"— -"]

        if len(extracted_texts) > 0:
            contract_texts = cls._clean_text(extracted_texts, start_page_patterns, end_page_patterns, middle_page_patterns)
        else:
            contract_texts = extracted_texts
            
        return contract_texts
    
    @classmethod
    def _extract_text_blocks(cls, base_str, contract_pattern, block_pattern, newline_pattern):
        matched_text = cls._row_list_regex(base_str, contract_pattern, block_pattern, newline_pattern)
        
        if matched_text != None:
            ext_blk_list = cls._mapped_positions_regex(matched_text)
            extracted_texts = cls._extract_texts_from_mapped_positions(ext_blk_list, base_str)
        else:
            extracted_texts = []

        return extracted_texts
    
    @classmethod
    def _clean_text(cls, ext_texts, start_page_patterns, end_page_patterns, middle_page_patterns):
        start_page_patterns = "|".join(start_page_patterns)
        middle_page_patterns = "|".join(middle_page_patterns)
        end_page_patterns = "|".join(end_page_patterns)

        page_patterns = [start_page_patterns, middle_page_patterns, end_page_patterns]
        page_patterns = "|".join(page_patterns)

        ext_texts = pd.Series(ext_texts).str.replace(page_patterns, "", regex=True)
        
        ext_texts_list = []
        
        for text in ext_texts:
            ext_texts_list.append(cls._remove_empty_line(text))
        
        return ext_texts_list
    
    @classmethod
    def _remove_empty_line(cls, text):
        return "\n".join([line for line in text.split("\n") if line.strip() != ""])

    @classmethod
    def _row_list_regex(cls, text, pattern_ext, pattern_blk, pattern_nl):
        # Lista com as buscas por regex. O findall permite ver se no documento há pelo menos 1 extrato detectado
        row_list = [re.findall(pattern_ext, text), re.finditer(pattern_ext, text), re.finditer(pattern_blk, text), re.finditer(pattern_nl, text)]

        # Se findall não achou nenhum, então o documento não tem nada que interessa
        if len(row_list[0]) == 0:
            return None
        
        return row_list

    @classmethod
    def _mapped_positions_regex(cls, matched_text):
        # Mapeia as posições do que foi encontrado pelo regex
        # Lista de 2 dimensões: na primeira, de extrato; na segunda, de bloco

        a_ext = matched_text[1]
        a_blk = matched_text[2]
        a_nl = matched_text[3]

        a_ext_list = []
        a_blk_list = []
        a_nl_list = []

        for i in a_ext:
            a_ext_list.append(i.start())

        for i in a_blk:
            a_blk_list.append(i.start())

        for i in a_nl:
            a_nl_list.append(i.start())

        return [a_ext_list, a_blk_list, a_nl_list]

    @classmethod
    def _extract_texts_from_mapped_positions(cls, mapped_positions, mapped_text):
        extracted_texts = []

        mapped_position = cls._extract_titles_blocks(mapped_positions)

        for ia in mapped_position[0]:
            ia_b = ia[0]
            ia_e = ia[-1]
            index_ia_b = mapped_position[1].index(ia_b)
            index_ia_e = mapped_position[1].index(ia_e)

            if (index_ia_e + 1) <= (len(mapped_position[1])-1):
                ib = mapped_position[1][index_ia_e+1]
                extracted_text = mapped_text[ia_b:ib]
            else:
                extracted_text = mapped_text[ia_b:]

            extracted_texts.append(extracted_text)

        return extracted_texts

    @classmethod
    def _extract_titles_blocks(cls, mapped_page):
        ext = mapped_page[0].copy()
        blk = mapped_page[1].copy()
        nl = mapped_page[2].copy()

        tls = []

        for t in ext:
            tl = []
            tl.append(t)

            loop_title = True

            # Um título acaba quando o próximo bloco não é a próxima linha
            while loop_title:
                prox_blk_index = blk.index(t)+1
                prox_nl_index = nl.index(t)+1

                # Se o fim do documento não tiver sido alcançado, então:
                if prox_blk_index < len(blk):
                    prox_blk = blk[prox_blk_index]
                    prox_nl = nl[prox_nl_index]

                    # Se a próxima linha for vista como um próximo bloco:
                    if prox_blk == prox_nl:
                        t = prox_blk
                        tl.append(t)
                    else:
                        loop_title = False

                else:
                    loop_title = False

            tls.append(tl)

        return [tls, blk, nl]


import pytest

from dodfminer.extract.pure.utils import title_filter


@pytest.fixture(name='filter_class', scope='module')
def fixture_filter_class():
    return title_filter.BoldUpperCase


def test_dict_text(filter_class):
    # Can be a title
    assert filter_class.dict_text({
        'text': "SECRETARIA DE PESSOAL DO DISTRITO FEDERAL",
    }) is True


def test_dict_text2(filter_class):
    # Can not be a title since has 4+ consecutive spaces
    assert filter_class.dict_text({
        'text': "SECRETARIA    DE PESSOAL DO DISTRITO FEDERAL",
    }) is False


def test_dict_text3(filter_class):
    # Can not be a title since has lower case chars
    assert filter_class.dict_text({
        'text': "SECRETARIA DE pessoal DO DISTRITO FEDERAL",
    }) is False


def test_dict_text4(filter_class):
    # Can not be a title since has 4- chars
    assert filter_class.dict_text({
        'text': "BOM",
    }) is False


def test_dict_bold(filter_class):
    assert filter_class.dict_bold({
        'flags': 14
    }) is False


def test_dict_bold2(filter_class):
    assert filter_class.dict_bold({
        'flags': 16
    }) is True


def test_dict_bold3(filter_class):
    assert filter_class.dict_bold({
        'flags': 18
    }) is False


def test_dict_bold4(filter_class):
    assert filter_class.dict_bold({
        'flags': 20
    }) is True



import re
import os
import joblib
from typing import List, Match
import pandas as pd
import numpy as np

from dodfminer.extract.polished.acts.base import Atos


def remove_crossed_words(s: str):
    """Any hyfen followed by 1+ spaces are removed.
    """
    return re.sub(r'-\s+', '', s)


LOWER_LETTER = r"[áàâäéèẽëíìîïóòôöúùûüça-z]"
UPPER_LETTER = r"[ÁÀÂÄÉÈẼËÍÌÎÏÓÒÔÖÚÙÛÜÇA-Z]"

DODF = r"(DODF|[Dd]i.rio\s+[Oo]ficial\s+[Dd]o\s+[Dd]istrito\s+[Ff]ederal)"

SIAPE = r"(?i:siape)\s*(?:n?.?)\s*(?P<siape>[-\d.Xx/\s]+)"
# SIAPE = r"(?i:siape)\s*(?:n?.?)\s*(?P<siape>[-\d.Xx/\s]+)"

MATRICULA = r"(?:matr.cul.|matr?[.]?\B)[^\d]+(?P<matricula>[-\d.XxZzYz/\s]+)"
# MATRICULA = r"(?i:matr.cul.|matr?[.]?\B)[^\d]+(?P<matricula>[-\d.XxZzYz/\s]+)"
MATRICULA_GENERICO = r"(?<![^\s])(?P<matricula>([-\d.XxZzYz/\s]{1,})[.-][\dXxYy][^\d])"
MATRICULA_ENTRE_VIRGULAS = r"(?<=[A-ZÀ-Ž]{3})\s*,\s+(?P<matricula>[-\d.XxZzYz/\s]{3,}?),"

SERVIDOR_NOME_COMPLETO = r"(servidor.?|empregad.)[^A-ZÀ-Ž]{0,40}(?P<name>[A-ZÀ-Ž][.'A-ZÀ-Ž\s]{6,}(?=[,]))"
# SERVIDOR_NOME_COMPLETO = r"(?i:servidor.?|empregad.)[^A-ZÀ-Ž]{0,40}(?P<name>[A-ZÀ-Ž][.'A-ZÀ-Ž\s]{6,}(?=[,]))"
NOME_COMPLETO = r"(?P<name>['A-ZÀ-Ž][.'A-ZÀ-Ž\s]{6,}(?=[,.:;]))"

PROCESSO_NUM = r"(?P<processo>[-0-9/.]+)"
INTERESSADO = r"(?i:interessad.):\s*{}".format(NOME_COMPLETO)
# INTERESSADO = r"(?i:interessad.):\s*{}".format(NOME_COMPLETO)
# INTERESSADO = r"(?i:interessad.):\s*" + NOME_COMPLETO

ONUS = r"(?P<onus>\b[oôOÔ](?i:nus)\b[^.]+[.])"
# ONUS = r"(?P<onus>\b[oôOÔ](?i:(nus))\b[^.]+[.])"


class Cessoes(Atos):
    _special_acts = ['matricula', 'cargo']

    def __init__(self, file, backend, debug=False, extra_search=True):
        self._debug = debug
        self._extra_search = extra_search
        # self._processed_text = remove_crossed_words(open(file).read())
        self._raw_matches = []
        super().__init__(file, backend)

    def _load_model(self):
        f_path = os.path.dirname(__file__)
        f_path += '/models/cessao.pkl'
        return joblib.load(f_path)

    def _load_seg_model(self):
        f_path = os.path.dirname(__file__)
        f_path += '/seg_models/cessao.pkl'
        return joblib.load(f_path)

    def _act_name(self):
        return "Cessoes"

    # def _load_model(self):
    #     f_path = os.path.dirname(__file__)
    #     f_path += '/models/cessoes_ner.pkl'
    #     return joblib.load(f_path)

    def _props_names(self):
        return ["tipo"] + list(self._prop_rules())

    def _rule_for_inst(self):
        return (
            r"([Pp][Rr][Oo][Cc][Ee][Ss][Ss][Oo][^0-9/]{0,12})([^\n]+?\n){0,2}?"
            + r"[^\n]*?[Aa]\s*[Ss]\s*[Ss]\s*[Uu]\s*[Nn]\s*[Tt]\s*[Oo]\s*:?\s*\bCESS.O\b"
            + r"([^\n]*\n){0,}?[^\n]*?(?=(?P<look_ahead>PROCESSO|Processo:|PUBLICAR|pertinentes[.]|autoridade cedente|"
            + r"(?i:publique-se)" + "))"
        )

    def _prop_rules(self):
        return {
            'nome': SERVIDOR_NOME_COMPLETO,
            'matricula': MATRICULA,
            'cargo_efetivo': "",
            'classe': "",
            'padrao': "",
            'orgao_cedente': "",
            'orgao_cessionario': "",
            'onus': ONUS,
            'fundamento legal': "",
            'processo_SEI': r"[^0-9]+?{}".format(PROCESSO_NUM),
            'vigencia': "",
            'matricula_SIAPE': SIAPE,
            'cargo_orgao_cessionario': r",(?P<cargo>[^,]+)",
            'simbolo': "",
            'hierarquia_lotacao': "",
        }

    def _find_instances(self) -> List[Match]:
        """Returns list of re.Match objects found on `self._text_no_crosswords`.

        Return:
            a list with all re.Match objects resulted from searching for
        """

        self._raw_matches = list(
            re.finditer(self._inst_rule, self._text, flags=self._flags)
            # Makes difficult to mark XML later despite improving results.
            # re.finditer(self._inst_rule, self._processed_text, flags=self._flags)
        )
        l = [i.group() for i in self._raw_matches]
        if self._debug:
            print("DEBUG:", len(l), 'matches')
        return l

    def _get_special_acts(self, lis_matches):
        for i, match in enumerate(self._raw_matches):
            act = match.group()
            matricula = re.search(MATRICULA, act) or \
                    re.search(MATRICULA_GENERICO, act) or \
                    re.search(MATRICULA_ENTRE_VIRGULAS, act)

            nome = re.search(self._rules['nome'], act)
            if matricula and nome:
                offset = matricula.end()-1 if 0 <= (matricula.start() - nome.end()) <= 5 \
                            else nome.end() - 1
                cargo, = self._find_prop_value(r",(?P<cargo>[^,]+)", act[offset:])
            else:
                cargo = np.nan

            lis_matches[i]['matricula'] = matricula.group('matricula') if matricula \
                                        else np.nan
            lis_matches[i]['cargo'] = cargo


    def _find_prop_value(self, rule, act):
        """Returns named group, or the whole match if no named groups
                are present on the match.
        Args:
            match: a re.Match object
        Returns: content of the unique named group found at match,
            the whole match if there are no groups at all or raise
            an exception if there are more than two groups.
        """
        match = re.search(rule, act, flags=self._flags)

        if match:
            keys = list(match.groupdict().keys())
            if len(keys) == 0:
                return match.group()
            elif len(keys) > 1:
                raise ValueError("Named regex must have AT MOST ONE NAMED GROUP.")
            if self._debug:
                print('key: ', keys[0])
            return (match.group(keys[0]),)
        else:
            return np.nan


    def _extract_props(self):
        acts = []
        for raw in self._raw_acts:
            act = self._regex_props(raw)
            acts.append(act)
        if self._extra_search:
            self._get_special_acts(acts)
        return acts


    def _regex_instances(self) -> List[Match]:
        found = self._find_instances()
        self._acts_str = found.copy()
        return found

    def _build_dataframe(self):
        """Create a dataframe with the extracted proprieties.

        Returns:
            The dataframe created
        """
        self._columns = self._prop_rules().keys()
        return (pd.DataFrame() if not self._acts else 
            pd.DataFrame(self._acts, columns = self._columns))



# coding=utf-8

"""Download DODFs from the Buriti Website and save on proper directory.

Download monthly pdfs of DODFs.

Usage example::

    downloader = Downloader()
    downloader.pull(start_date, end_date)

"""

import os
from pathlib import Path
from datetime import datetime
import tqdm
import requests

from dateutil.relativedelta import relativedelta
from dodfminer.downloader.helper import check_date, get_downloads


MONTHS_STRING = ["", "01_Janeiro", "02_Fevereiro", "03_Março", "04_Abril",
                 "05_Maio", "06_Junho", "07_Julho", "08_Agosto",
                 "09_Setembro", "10_Outubro", "11_Novembro", "12_Dezembro"]


class Downloader:
    """Responsible for the download of the DODFs Pdfs.

    Args:
        save_path (str): Path to save the downloads.

    Attributes:
        _download_path: Folder in which the downloads will be stored.
        _prog_bar: Indicate if download should contain a progress bar.

    """

    def __init__(self, save_path='./'):
        self._prog_bar = tqdm.tqdm()
        self._create_single_folder(os.path.join(save_path, 'dodfs'))
        self._download_path = os.path.join(save_path, 'dodfs')

    @classmethod
    def _string_to_date(cls, date):
        """Convert the date to datetime.

        Args:
            date (:obj:`datetime`): The date to be converted in string format.

        Returns:
            Return the date formated in string now as datetime datatype.

        Raises:
            Exception: date passed through cli is in wrong format.

        """
        if '/' in date:
            date = datetime.strptime(date, '%m/%Y').date()
        elif '-' in date:
            date = datetime.strptime(date, '%m-%Y').date()
        else:
            msg = 'start_date or end_date must be in format mm/yyyy or mm-yyyy'
            raise Exception(msg)

        return date

    def _create_single_folder(self, path):
        """Create a single folder given the directory path.

        This function might create a folder, observe that the folder already
        exists, or raise an error if the folder cannot be created.

        Args:
            path (str): The path to be created

        Raises:
            OSError: Error creating the directory.

        """
        if os.path.exists(path):
            self._log(os.path.basename(path) + " folder already exist")
        else:
            try:
                os.mkdir(path)
            except OSError as error:
                self._log("Exception during the directory creation")
                self._log(str(error))
                raise
            else:
                basename = os.path.basename(path)
                self._log(basename + " directory successful created")

    def _create_download_folder(self):
        """Create Downloaded DODFs Structures."""
        # import pdb; pdb.set_trace()
        self._create_single_folder(self._download_path)

    def _fail_request_message(self, url, error):
        """Log error messages in download.

        Args:
            url (str): The failing url to the website.
            error (str): The kind of error happening.

        """
        self._log(error)
        message = "Please check your internet connection, and " \
                  f"check if the url is online via browser: {url}"
        self._log(message)

    def _file_exist(self, path):
        """Check if a file exists.

        Prevents redownloads.

        Args:
            path (str): The path where the file might be

        Returns:
            Boolean indicating if file does really exists.

        """
        if os.path.exists(path):
            self._log(os.path.basename(path) + " file already exist")
            return True

        return False

    def _download_pdf(self, url, path):
        """Download the DODF PDF.

        Note:
            Might be time consuming depending on bandwidth.

        Args:
            url (str): The pdf url.
            path (str): The path to save the pdf.

        Raises:
            RequestException: Error in case the request to download fails.

        """
        try:
            response = requests.get(url)
            response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            self._fail_request_message(url, error)
        except requests.exceptions.RequestException as error:
            self._fail_request_message(url, error)
        else:
            pdf_file = Path(f"{path}.pdf")
            pdf_file.write_bytes(response.content)
            self._log("Finished " + os.path.basename(path))

    def _make_month_path(self, year, actual_date):
        """Create and return the folder for the year and month being download.

        Args:
            year (int): The year respective to the folder.
            actual_date (:obj:`datetime`): The date in which the downloaded
            DODF corresponds.

        Returns:
            The path to the actual month in which the download is being made.

        """
        year_path = os.path.join(self._download_path,
                                 str(actual_date.year))
        if year != actual_date.year:
            self._create_single_folder(year_path)
        month_path = os.path.join(year_path, MONTHS_STRING[actual_date.month])

        return month_path

    def pull(self, start_date, end_date):
        """Make the download of the DODFs pdfs.

        All dodfs are downloaded from start_date to end_date inclusively.
        The Pdfs are saved in a folder called "data" inside the project folder.

        Args:
            start_date (str): The start date in format mm/yyyy.
            end_date (str): The start date in format mm/yyyy.

        Note:
            The name or the path of the save folder are hard coded and can't
            be changed due to some nonsense software engineer decision.

        """
        # Convert string to datetime and calculate ammount to be used in
        # progress bar
        start_date = self._string_to_date(start_date)
        end_date = self._string_to_date(end_date)
        months_amt = ((end_date.year - start_date.year) * 12
                      + (end_date.month - start_date.month))
        # Creates progress bar
        self._prog_bar = tqdm.tqdm(total=months_amt)
        # # Creates the project folder structure
        self._create_download_folder()
        year = 0

        for month in range(months_amt+1):
            actual_date = start_date + relativedelta(months=+month)
            desc_bar = str(actual_date)
            self._prog_bar.set_description(f"Date {desc_bar}")
            month_path = self._make_month_path(year, actual_date)
            year = actual_date.year
            year_ = str(year)
            month_ = MONTHS_STRING[actual_date.month]

            if check_date(year_, month_) is True:
                self._create_single_folder(month_path)
            else:
                print(
                    f"*** There are still no DODFs for that date: {actual_date.month}/{year_} ***")
                continue

            self._get_dodfs(get_downloads(year_, month_), month_path)

        self._prog_bar.update(1)


    def _get_dodfs(self, _links_for_each_dodf, month_path):
        """Create folder and stores the DODFs pdfs.

        Args:
            _links_for_each_dodf (dict): a dicts with links for each DODF.
            month_path (str): path to store DODFs pdfs.

        """
        for dodf_name, links in _links_for_each_dodf.items():
            dodf_path = month_path

            if len(links) > 1:
                dodf_path = os.path.join(month_path, dodf_name)
                self._create_single_folder(dodf_path)

            index = 0
            for link in links:
                index += 1
                download_link = link
                if len(links) == 1:
                    dodf_name_path = os.path.join(dodf_path, dodf_name)
                else:
                    dodf_name_path = os.path.join(
                        dodf_path, f'{dodf_name} {index}')

                if not self._file_exist(dodf_name_path):
                    self._log("Downloding " +
                                os.path.basename(dodf_name_path))
                    self._download_pdf(download_link, dodf_name_path)
                else:
                    self._log("Jumping to the next")

    def get_download_path(self):
        return self._download_path

    def _log(self, message):
        """Logs a message following the downloader pattern.

        Args:
            message (str): The message to be logged.

        """
        self._prog_bar.write("[DOWNLOADER] " + str(message))


if __name__ == '__main__':
    downloader = Downloader(save_path='./')
    downloader.pull(start_date="05/2021", end_date="06/2021")



"""Committee Classification Class.

This module contains the Committee class, with all the necessary
functions for preprocessing and classification with a committee
of classifiers.
"""

import nltk
import numpy as np
import joblib

from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import FunctionTransformer

class Committee:
    """Committee class.
    
    Creates a pipeline for preprocessing and classification when initialized.

    Args:
        path (str): The path to the models file.

    The models file needs to contain a list of classifiers and an object
    with the transform function to be used as embedding.
    """
    def __init__(self, path):
        self.pipe = make_pipeline(FunctionTransformer(self.remove_not_words),
                             FunctionTransformer(self.embedding),
                             FunctionTransformer(self.predict),
                             FunctionTransformer(self.get_new_labels))
        
        nltk.download('punkt', quiet = True)
        nltk.download('rslp', quiet = True)
        nltk.download('stopwords', quiet = True)
        
        self.stopwords = nltk.corpus.stopwords.words('portuguese')
        self.stemmer = nltk.stem.RSLPStemmer()

        self.load_models(path)

    def load_models(self, path):
        """ Loads a list of classifiers and the object used for embedding from the specified path. """
        with open(path, 'rb') as file:
            self.clfs = joblib.load(file)
            self.vectorizer = joblib.load(file)
            self.vectorizer.tokenizer = self.tokenize
        
    def transform(self, X, y):
        """ Initializes the pipeline for preprocessing and classification. """
        self.y = y

        if len(y) == 0:
            return []
        
        return self.pipe.transform(X)
    
    def tokenize(self, text):
        """ Used by the vectorizer to tokenize text. """
        tokens = [word.lower() for word in nltk.word_tokenize(text) if len(word) > 1]
        return [self.stemmer.stem(item) for item in tokens if item not in self.stopwords]

    def remove_not_words(self, text):
        """ Removes everything that is not a word from a text. """
        return text.str.replace(r'\w*\d\w*|\W', ' ', regex = True).str.lower()

    def embedding(self, text):
        """ Transforms the data with an embedding function. """
        return self.vectorizer.transform(text)

    def predict(self, text):
        """ Generates classifiers predictions for the data. """
        results = []
        for clf in self.clfs:
            predicted = clf.predict(text)
            results.append(predicted)

        return np.array(results).T
        
    def get_new_labels(self, results):
        """ Uses a committee to decide new labels from the classifiers predictions.
        
        If the committee does not agree on any label for an act,
        the original label is maintained. 
        """
        new_types = []
        for i in range(0, len(results)):
            d = dict(zip(*np.unique(results[i], return_counts=True)))
            r = max(d, key = d.get)
            if d[r] >= 0.5 and r != self.y[i]:
                new_types.append(r)
            else:
                new_types.append(self.y[i])
        return new_types


import os
import pytest

import pandas as pd
from pandas.core.frame import DataFrame

from tests.helpers.decorators import clean_extra_files

from dodfminer.extract.pure.core import ContentExtractor
from dodfminer.extract.polished.helper import xml_multiple, get_files_path, build_act_txt, extract_single, extract_multiple, \
    extract_multiple_acts, extract_multiple_acts_with_committee, committee_classification


FOLDER_PATH = f"{os.path.dirname(__file__)}/support/polished"

###################### FIXTURES ######################


@pytest.fixture(name="folder_path")
def fixture_folder_path():
    return FOLDER_PATH


@pytest.fixture(name="file_path")
def fixture_file_path(folder_path):
    def _file_path(extension="pdf"):
        return f"{folder_path}/DODF 001 01-01-2019 EDICAO ESPECIAL.{extension}"
    return _file_path


###################### TESTS ######################
@clean_extra_files(FOLDER_PATH)
def test_helper_xml_multiple(folder_path):
    try:
        xml_multiple(folder_path, "regex")
        xml_files_list = list(
            filter(lambda x: ".xml" in x, os.listdir(folder_path)))

        assert "1_1.1.2019.xml" in xml_files_list
    except AssertionError:
        assert False


@clean_extra_files(FOLDER_PATH)
def test_helper_extract_multiple(folder_path):
    ContentExtractor.extract_to_txt(folder_path)
    files = get_files_path(f"{folder_path}/results/txt", 'txt')
    data_frame = extract_multiple(
        files, "nomeacao", "regex", txt_path="./results")

    assert len(data_frame) > 0


@clean_extra_files(FOLDER_PATH)
def test_helper_extract_single(file_path):
    ContentExtractor.extract_text(file_path(extension="pdf"), single=True)
    data_frame, texts = extract_single(
        file_path(extension="txt"), "nomeacao", "regex")

    assert isinstance(data_frame, DataFrame)
    assert isinstance(texts, list)

    assert len(data_frame) > 0
    assert len(texts) > 0


@clean_extra_files(FOLDER_PATH)
def test_helper_extract_multiple_acts(folder_path, file_path):
    extract_multiple_acts(folder_path, ["nomeacao"], "regex")
    multiple_files_df = pd.read_csv(f"{folder_path}/nomeacao.csv")
    extract_multiple_acts(file_path(extension="pdf"), ["nomeacao"], "regex")
    single_file_df = pd.read_csv(f"{folder_path}/nomeacao.csv")

    assert len(multiple_files_df) > 0
    assert len(single_file_df) > 0
    assert len(multiple_files_df) > len(single_file_df)


@clean_extra_files(FOLDER_PATH)
def test_helper_extract_multiple_acts_with_committee(folder_path, file_path):
    extract_multiple_acts_with_committee(folder_path, ["nomeacao"], "regex")
    multiple_files_df = pd.read_csv(f"{folder_path}/nomeacao.csv")

    extract_multiple_acts_with_committee(file_path(extension="pdf"), ["nomeacao"], "regex")
    single_file_df = pd.read_csv(f"{folder_path}/nomeacao.csv")

    assert len(multiple_files_df) > 0
    assert len(single_file_df) > 0
    assert len(multiple_files_df) > len(single_file_df)

@clean_extra_files(FOLDER_PATH)
def test_helper_committee_classification(folder_path, file_path):
    dataframe = pd.read_csv(f"{folder_path}/nomeacao_extraida.csv")
    dataframe = dataframe.filter(['Tipo do Ato', 'text'], axis = 1)
    dataframe.columns = ['type', 'text']

    committee_classification(dataframe, file_path(extension="pdf"), ["nomeacao"], "regex")
    output_df = pd.read_csv(f"{folder_path}/nomeacao.csv")

    assert len(output_df) == len(dataframe)

@clean_extra_files(FOLDER_PATH)
def test_helper_build_act_txt():
    directory = ""+os.path.dirname(__file__)+"/support/"
    build_act_txt(["aposentadoria"], "crioutxt", save_path=directory)
    assert "crioutxt.txt" in os.listdir(directory)
    os.remove(os.path.join(directory, "crioutxt.txt"))

# def test_helper_print_dataframe():
#     df = print_dataframe(pd.DataFrame())
#     assert isinstance(df, pd.io.formats.style.Styler)


def test_helper_get_files_path(folder_path):
    files = get_files_path(folder_path, "txt")
    assert len(files) > 0
    files = get_files_path(folder_path, "pdf")
    assert len(files) > 0



# pylint: disable=protected-access

import os
import json
from pathlib import Path
import pytest
import fitz

from dodfminer.extract.pure.utils import box_extractor

BASE_PATH = Path(
    ""+os.path.dirname(__file__)+"/support"
)
PDF_PATH = BASE_PATH/'03-12-2018.pdf'
PAGE_ONE_LINES_PATH = BASE_PATH/'03-12-2018_pagina1_lines.json'
PAGE_49_LINES_PATH = BASE_PATH/'03-12-2018_pagina49_lines.json'
GET_DOC_TEXT_BOXES_PATH = BASE_PATH/'03-12-2018_get_doc_text_boxes.json'
GET_DOC_TEXT_LINES_PATH = BASE_PATH/'03-12-2018_get_doc_text_lines.json'
GET_DOC_IMG_PATH = BASE_PATH/'03-12-2018_get_doc_img.json'
GET_DOC_IMG_BOXES_PATH = BASE_PATH/'03-12-2018_get_doc_img_boxes.json'


def tuplefy(lis):
    """Turn iterables into tuples. Makes comparison easier."""
    try:
        return tuple(tuplefy(i) for i in lis)
    except TypeError:
        return tuple(tuple(i) for i in lis)


@pytest.fixture(scope='module', name='pdf_fitz')
def fixture_pdf_fitz():
    return fitz.open(PDF_PATH.as_posix())


def wrapper_extract_page_lines_content(pdf_fitz, page_num, path):
    # Can be a title
    as_tup = tuplefy(
        box_extractor._extract_page_lines_content(pdf_fitz[page_num])
    )
    with open(path, encoding='utf-8') as json_file:
        ground_truth = json.load(json_file)
        ground_truth = tuplefy(ground_truth)
        assert as_tup == ground_truth


def test_extract_page_lines_content_1(pdf_fitz):
    # Can be a title
    wrapper_extract_page_lines_content(
        pdf_fitz, 0, PAGE_ONE_LINES_PATH
    )


def test_extract_page_lines_content_49(pdf_fitz):
    # Can be a title
    wrapper_extract_page_lines_content(
        pdf_fitz, 48, PAGE_49_LINES_PATH
    )


def test_get_doc_text_boxes(pdf_fitz):
    assert len(box_extractor.get_doc_text_boxes(pdf_fitz)) == len(pdf_fitz)


def test_doc_text_lines(pdf_fitz):
    with open(GET_DOC_TEXT_LINES_PATH.as_posix(), encoding='utf-8') as json_file:
        ground_truth = json.load(json_file)
        assert (
            tuplefy(ground_truth)
            == tuplefy(box_extractor.get_doc_text_lines(pdf_fitz))
        )


def test_get_doc_img(pdf_fitz):
    with open(GET_DOC_IMG_PATH.as_posix(), encoding='utf-8') as json_file:
        ground_truth = json.load(json_file)
        assert (
            tuplefy(ground_truth)
            == tuplefy(box_extractor._get_doc_img(pdf_fitz))
        )


def test_get_doc_img_boxes(pdf_fitz):
    with open(GET_DOC_IMG_BOXES_PATH.as_posix(), encoding='utf-8') as json_file:
        ground_truth = json.load(json_file)
        assert (
            tuplefy(ground_truth)
            == tuplefy(box_extractor.get_doc_img_boxes(pdf_fitz))
        )



import shutil
import pytest
import requests
from dodfminer.downloader.core import Downloader
from dodfminer.downloader.helper import LISTAR_URL


def test_download_date_fail():
    with pytest.raises(Exception):
        downloader = Downloader()
        downloader.pull(start_date="05\\2021", end_date="1006\\2021")


def test_download_date_fail_2(requests_mock):
    requests_mock.get(
        f'{LISTAR_URL}dir=2021/05_Maio', json={})
    requests_mock.get(
        f'{LISTAR_URL}dir=2021/06_Junho', json={})
    downloader = Downloader()
    downloader.pull(start_date="05/2021", end_date="06/2021")


def test_downloader_fail(requests_mock):

    requests_mock.get(f'{LISTAR_URL}dir=2021/05_Maio', json={
        "data": {
            "20170131100": "DODF 022 31-01-2017",
        }
    })
    requests_mock.get(f'{LISTAR_URL}dir=2021/05_Maio/DODF%20022%2031-01-2017', json={
        "data": {
            "UNIT": "Test",
        }
    })
    requests_mock.get(f'{LISTAR_URL}dir=2021/05_Maio/DODF%20003%2030-01-2017%20EDICAO EXTRA', json={
        "data": {
            "UNIT": "Test2",
        }
    })

    requests_mock.register_uri('GET',
                               'https://dodf.df.gov.br/index/visualizar-arquivo/?pasta=2021%7C05_Maio%7CDODF%20022%2031-01-2017%7C&arquivo=UNIT',
                               exc=requests.exceptions.HTTPError)
    downloader = Downloader()
    downloader.pull(start_date="05-2021", end_date="05-2021")


def test_downloader(requests_mock):
    # Remove conteudo inicial
    shutil.rmtree('dodfs/2021', ignore_errors=True)

    requests_mock.get(f'{LISTAR_URL}dir=2021/05_Maio', json={
        "data": {
            "20170131100": "DODF 022 31-01-2017",
            "20170130103": "DODF 003 30-01-2017 EDICAO EXTRA",
        }
    })
    requests_mock.get(f'{LISTAR_URL}dir=2021/05_Maio/DODF%20022%2031-01-2017', json={
        "data": {
            "UNIT": "Test",
        }
    })
    requests_mock.get(f'{LISTAR_URL}dir=2021/05_Maio/DODF%20003%2030-01-2017%20EDICAO EXTRA', json={
        "data": {
            "UNIT": "Test2",
        }
    })

    requests_mock.get(
        'https://dodf.df.gov.br/index/visualizar-arquivo/?pasta=2021%7C05_Maio%7CDODF%20022%2031-01-2017%7C&arquivo=UNIT', json={})
    requests_mock.get(
        'https://dodf.df.gov.br/index/visualizar-arquivo/?pasta=2021%7C05_Maio%7CDODF%20003%2030-01-2017%20EDICAO%20EXTRA%7C&arquivo=UNIT', json={})

    requests_mock.get(f'{LISTAR_URL}dir=2021/06_Junho', json={
        "data": {
            "20170131100": "DODF 022 31-01-2017",
            "20170130103": "DODF 003 30-01-2017 EDICAO EXTRA",
        }
    })
    requests_mock.get(f'{LISTAR_URL}dir=2021/06_Junho/DODF%20022%2031-01-2017', json={
        "data": {
            "UNIT": "Test",
        }
    })
    requests_mock.get(f'{LISTAR_URL}dir=2021/06_Junho/DODF%20003%2030-01-2017%20EDICAO EXTRA', json={
        "data": {
            "UNIT": "Test2",
        }
    })

    requests_mock.get(
        'https://dodf.df.gov.br/index/visualizar-arquivo/?pasta=2021%7C06_Junho%7CDODF%20022%2031-01-2017%7C&arquivo=UNIT', json={})
    requests_mock.get(
        'https://dodf.df.gov.br/index/visualizar-arquivo/?pasta=2021%7C06_Junho%7CDODF%20003%2030-01-2017%20EDICAO%20EXTRA%7C&arquivo=UNIT', json={})

    downloader = Downloader()
    downloader.pull(start_date="05/2021", end_date="06/2021")


def test_downloader_2(requests_mock):
    requests_mock.get(f'{LISTAR_URL}dir=2021/05_Maio', json={
        "data": {
            "20170131100": "DODF 022 31-01-2017",
        }
    })
    requests_mock.get(f'{LISTAR_URL}dir=2021/05_Maio/DODF%20022%2031-01-2017', json={
        "data": {
            "UNIT": "Test",
        }
    })
    requests_mock.get(f'{LISTAR_URL}dir=2021/05_Maio/DODF%20003%2030-01-2017%20EDICAO EXTRA', json={
        "data": {
            "UNIT": "Test2",
        }
    })

    requests_mock.get(
        'https://dodf.df.gov.br/index/visualizar-arquivo/?pasta=2021%7C05_Maio%7CDODF%20022%2031-01-2017%7C&arquivo=UNIT', json={})
    downloader = Downloader()
    downloader.pull(start_date="05-2021", end_date="05-2021")



"""
Pure extractor core module.

This module contains the ActsExtractor class, which have all that is necessary to
extract a single act or all the acts from a DODF.

Usage Example::

    from dodfminer.extract.polished.core import ActsExtractor
    ActsExtractor.get_act_obj(ato_id, file, backend)

"""

from dodfminer.extract.polished.acts.aposentadoria import Retirements, RetAposentadoria
from dodfminer.extract.polished.acts.nomeacao import NomeacaoComissionados, NomeacaoEfetivos
from dodfminer.extract.polished.acts.exoneracao import Exoneracao, ExoneracaoEfetivos
from dodfminer.extract.polished.acts.reversoes import Revertions
from dodfminer.extract.polished.acts.abono import AbonoPermanencia
from dodfminer.extract.polished.acts.substituicao import Substituicao
from dodfminer.extract.polished.acts.cessoes import Cessoes
from dodfminer.extract.polished.acts.sem_efeito_aposentadoria import SemEfeitoAposentadoria
from dodfminer.extract.polished.acts.contrato import Contratos
from dodfminer.extract.polished.create_xml import XMLFy

_acts_ids = {"aposentadoria": Retirements,
             "reversoes": Revertions,
             "nomeacao": NomeacaoComissionados,
             "exoneracao": Exoneracao,
             "abono": AbonoPermanencia,
             "retificacoes": RetAposentadoria,
             "substituicao": Substituicao,
             "efetivos_nome": NomeacaoEfetivos,
             "efetivos_exo": ExoneracaoEfetivos,
             "sem_efeito_aposentadoria": SemEfeitoAposentadoria,
             "cessoes": Cessoes,
             "contrato": Contratos}

"""_acts_ids: All avaiable acts classes indexed by a given string name."""


class ActsExtractor:
    """Polished Extraction main class.

    All interactions with the acts needs to be done through this interface.
    This class handles all the requests to Regex or NER extraction.

    Note:
        This class is static.

    """

    @staticmethod
    def get_act_obj(ato_id, file, backend):
        """
        Extract a single act type from a single DODF.

        Object format.

        Args:
            ato_id (string): The name of the act to extract.
            file (string): Path of the file.
            backend (string): Backend of act extraction, either Regex or NER.

        Returns:
            An object of the desired act, already with extracted information.

        """
        return _acts_ids[ato_id](file, backend)

    @staticmethod
    def get_all_obj(file, backend):
        """
        Extract all act types from a single DODF object.

        Object format.

        Args:
            file (string): Path of the file.
            backend (string): Backend of act extraction, either Regex or NER.

        Returns:
            An vector of objects of all the acts with extracted
            information.

        """
        res = {}
        for key, act in _acts_ids.items():
            res[key] = act(file, backend)

        return res

    @staticmethod
    def get_act_df(ato_id, file, backend):
        """
        Extract a single act type from a single DODF.

        Dataframe format.

        Args:
            ato_id (string): The name of the act to extract.
            file (string): Path of the file.
            backend (string): Backend of act extraction, either Regex or NER.

        Returns:
            A dataframe with extracted information, for the desired act.

        """
        return _acts_ids[ato_id](file, backend).data_frame

    @staticmethod
    def get_all_df(file, backend):
        """
        Extract all act types from a single DODF file.

        Dataframe format.

        Args:
            file (string): Path of the file.
            backend (string): Backend of act extraction, either regex or ner.

        Returns:
            A vector of dataframes with extracted information for all acts.

        """
        res = {}
        for key, act in _acts_ids.items():
            res[key] = act(file, backend).data_frame

        return res

    @staticmethod
    def get_xml(file, _, i):
        """
        Extract all act types from a single DODF in xml.

        Dataframe format.

        Args:
            file (string): Path of the file.
            backend (string): Backend of act extraction, either regex or ner.

        Returns:
            A vector of dataframes with extracted information for all acts.

        """
        res = XMLFy(file, _acts_ids, i)
        return res



import json
import requests


LISTAR_URL = 'https://dodf.df.gov.br/listar?'
DOWNLOAD_URL = 'https://dodf.df.gov.br/index/visualizar-arquivo/?pasta='


# req1 returns list with available months for the given year
def req1(year):
    url1 = LISTAR_URL+f'dir={str(year)}'
    return url1


# req2 returns list with available DODFs for the given month
def req2(url1, month):
    if "_" in month:
        url2 = url1 + '/' + month
        request2 = requests.get(url2)
        content2 = json.loads(request2.content)
        dodfs = list(content2['data'].items())
        return(url2, dodfs)

    raise ValueError("month parameter format is wrong")


# req3 returns all pdfs from the selected DODF
def req3(url2, dodf):
    url3 = url2+"/"+dodf.replace(" ", "%20")
    request3 = requests.get(url3)
    content3 = json.loads(request3.content)
    pdfs = content3['data']
    return(url3, pdfs)


# Check for DODFs on the selected date
def check_date(year, month):
    if "_" in month:
        url1 = LISTAR_URL+f'dir={year}/{month}'
        print(url1)
        req = requests.get(url1)
        content = json.loads(req.content)
        if 'data' in content.keys():
            return len(content["data"]) > 0
        return False

    raise ValueError("month parameter format is wrong")

# Generates download url


def get_downloads(year, month):
    if "_" in month:
        url2, dodfs = req2(req1(year), month)
        _links = {}

        # Lista de DODFS:
        for dodf in dodfs:
            url3, pdfs = req3(url2, dodf[1])
            dodf_name = dodf[1]
            _pdfs = []
            for pdf in pdfs:
                dir_ = (url3[url3.find(year):]).replace("/", "|")+'|'+'&arquivo='
                index2 = pdf.rfind("/")
                arq_ = pdf[index2+1:].replace(" ", "%20")
                link_download = DOWNLOAD_URL+dir_+arq_
                _pdfs.append(link_download)
                _links[dodf_name] = _pdfs

        return _links

    raise ValueError("month parameter format is wrong")



import pytest
from dodfminer.downloader.helper import req1, req2, req3, get_downloads, check_date, LISTAR_URL


def test_req1():
    assert req1(2017) == f'{LISTAR_URL}dir=2017'


def test_req2(requests_mock):
    # Cria um mock do retorno da requisicao
    requests_mock.get(f'{LISTAR_URL}dir=2017/01_Janeiro', json={
        "data": {
            "20170131100": "DODF 022 31-01-2017",
            "20170130103": "DODF 003 30-01-2017 EDICAO EXTRA",
        }
    })
    assert isinstance(
        req2(f'{LISTAR_URL}dir=2017', '01_Janeiro'), tuple)


def test_req2_2():
    with pytest.raises(ValueError):
        assert req2(f'{LISTAR_URL}dir=2017', 'Janeiro')


def test_req3(requests_mock):
    requests_mock.get(f'{LISTAR_URL}dir=2017/DODF%20023%2031-01-2006', json={
        "data": {
            "20170131100": "DODF 022 31-01-2017",
            "20170130103": "DODF 003 30-01-2017 EDICAO EXTRA",
        }
    })
    assert isinstance(req3(f'{LISTAR_URL}dir=2017',
                'DODF 023 31-01-2006'), tuple)


def test_get_downloads(requests_mock):
    # Primeira requisicao
    requests_mock.get(f'{LISTAR_URL}dir=2017/01_Janeiro', json={
        "data": {
            "20170131100": "DODF 022 31-01-2017",
            "20170130103": "DODF 003 30-01-2017 EDICAO EXTRA",
        }
    })
    # Segunda requisicao
    requests_mock.get(f'{LISTAR_URL}dir=2017/01_Janeiro/DODF%20022%2031-01-2017', json={
        "data": {
            "UNIT": "Test",
        }
    })
    # Terceira requisicao
    requests_mock.get(f'{LISTAR_URL}dir=2017/01_Janeiro/DODF%20003%2030-01-2017%20EDICAO EXTRA', json={
        "data": {
            "UNIT": "Test2",
        }
    })

    assert isinstance(get_downloads('2017', '01_Janeiro'), dict)


def test_get_downloads_2():
    with pytest.raises(ValueError):
        assert get_downloads('2017', 'Janeiro')


def test_check_date(requests_mock):
    requests_mock.get(f'{LISTAR_URL}dir=2017/01_Janeiro', json={
        "data": {
            "Unit": "Test"
        }
    })
    # Deve retornar true
    assert check_date('2017', '01_Janeiro')

    requests_mock.get(f'{LISTAR_URL}dir=2017/01_Janeiro', json={
        "data": {}
    })
    # Deve retornar false
    assert not check_date('2017', '01_Janeiro')

    requests_mock.get(
        f'{LISTAR_URL}dir=2017/01_Janeiro', json={})
    # Deve retornar false
    assert not check_date('2017', '01_Janeiro')


def test_check_date_2():
    with pytest.raises(ValueError):
        assert check_date('2017', 'Janeiro')



import os
import shutil
import functools


def clean_extra_files(folder_path):
    def _clean_extra_files(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            original_file_system_entries = os.listdir(folder_path)
            func(*args, **kwargs)
            final_file_system_entries = os.listdir(folder_path)

            extra_file_system_entries = filter_extra_files(
                original_file_system_entries, final_file_system_entries)
            for entry in extra_file_system_entries:
                delete_file_or_folder(entry, folder_path)

        return wrapper
    return _clean_extra_files


def filter_extra_files(original, final):
    def is_not_original(final_entries):
        return final_entries not in original
    return list(filter(is_not_original, final))


def delete_file_or_folder(entry, folder_path):
    if os.path.isfile(os.path.join(folder_path, entry)):
        os.remove(os.path.join(folder_path, entry))
    else:
        shutil.rmtree(os.path.join(folder_path, entry), ignore_errors=True)



version_info = (1, 3, 9)
# format:
# ('dodf_major', 'dodf_minor', 'dodf_patch')


def get_version():
    "Returns the version as a human-format string."
    return f'{version_info[0]}.{version_info[1]}.{version_info[2]}'


__version__ = get_version()



import os
import datetime
import re
from lxml import etree
from dodfminer.extract.pure.core import ContentExtractor as ce

class XMLFy:
    '''Cria um xml com informações de um ato'''

    def __init__(self, file, acts_ids, id_xml):
        self._file = file
        self._acts_ids = acts_ids
        self._xml_id = self.build_xml_id(id_xml)
        self._annotation_id = 1
        self._relations_id = 1
        self.xml = self._create_xml()

    def build_xml_id(self, id_xml):
        file_name = self._file.split('/')[-1]

        str2int2str = lambda x : str(int(x))
        file_numbers_list = map(str2int2str, re.findall(r'\d+', file_name))

        file_id = ".".join(list(file_numbers_list)[1:])

        return f"{id_xml}_{file_id}"

    def print_tree(self):
        print(etree.tostring(self.xml, pretty_print=True).decode())

    def save_to_disc(self, path):
        doc_type = """<!DOCTYPE collection SYSTEM "BioC.dtd">"""
        if os.path.isfile(path):
            folder_path = os.path.splitext(path)[0]
            print(folder_path)
        else:
            folder_path = path
        self.xml.write(os.path.join(folder_path, self._xml_id+".xml"), pretty_print=True, xml_declaration=True,
                      encoding="utf-8", doctype=doc_type)


    def _create_xml(self):
        doc = self._creat_collection()
        tree = etree.ElementTree(doc)
        return tree


    def _creat_collection(self):
        root_collect = etree.Element('collection')
        root_collect.append(etree.Element('source'))
        root_collect.append(etree.Element('date'))
        root_collect.append(etree.Element('key'))

        doc = self._create_document()

        root_collect.append(doc)

        return root_collect


    def _create_document(self):
        root_doc = etree.Element('document')

        child_id = etree.Element('id')
        child_id.text = str(self._xml_id)
        root_doc.append(child_id)

        infon = etree.Element('infon')
        infon.set("key", "tt_curatable")
        infon.text = "no"
        root_doc.append(infon)

        infon = etree.Element('infon')
        infon.set("key", "tt_version")
        infon.text = "0"
        root_doc.append(infon)

        infon = etree.Element('infon')
        infon.set("key", "tt_round")
        infon.text = "0"
        root_doc.append(infon)

        self._text_to_passages(root_doc)

        return root_doc

    def _text_to_passages(self, root):
        text = ce.extract_text(self._file, single=False,
                                  block=True, sep=' ', norm='NFKD')
        offset = 0
        for line in text:
            _, _, _, _, text = line
            annotation = self.execute_regex(text)
            child = self._create_passage(offset, text, annotation)
            root.append(child)
            offset += len(text)-1

    # pylint: disable=protected-access
    def execute_regex(self, text):
        res = {}
        for key in self._acts_ids:
            act = self._acts_ids[key](text, "regex")
            if act._acts:
                res[key] = act._acts
        return res

    def _create_passage(self, offset, text, an_dict):
        root_passage = etree.Element('passage')

        child_offset = etree.Element('offset')
        child_offset.text = str(offset)
        root_passage.append(child_offset)

        child_text = etree.Element('text')
        child_text.text = text
        root_passage.append(child_text)

        # relations = []
        # if an_dict:
        for act_type in an_dict:
            for act_inst in an_dict[act_type]:
                # relations_ids = []
                for prop_type, value in act_inst.items():
                    if isinstance(value, str):
                        offset_sum = text.find(value)
                        text = text.replace(value, '_'*len(value), 1)
                        if offset_sum != -1:
                            offset_prop = offset + offset_sum
                            child_annotation = self._annotate(prop_type,
                                                                value, offset_prop)
                            root_passage.append(child_annotation)
                            # relations_ids.append(self._annotation_id)
                            self._annotation_id += 1
                # child_rel = self._create_relation("Ato_"+act_type.capitalize(), relations_ids)
                # relations.append(child_rel)
                # self._relations_id += 1

        # for relation in relations:
        #     root_passage.append(relation)

        return root_passage

    def _annotate(self, annotation_type, text, offset):
        root_annotate = etree.Element('annotation')
        root_annotate.set("id", str(self._annotation_id))

        infon = etree.Element('infon')
        infon.set("key", "type")
        infon.text = annotation_type
        root_annotate.append(infon)

        infon = etree.Element('infon')
        infon.set("key", "identifier")
        root_annotate.append(infon)

        infon = etree.Element('infon')
        infon.set("key", "annotator")
        infon.text = "DODFMiner"
        root_annotate.append(infon)

        infon = etree.Element('infon')
        infon.set("key", "updated_at")
        infon.text = str(datetime.datetime.now())
        root_annotate.append(infon)

        location = etree.Element('location')
        location.set("offset", str(offset))
        location.set("length", str(len(text)))
        root_annotate.append(location)

        annotation_txt = etree.Element('text')
        annotation_txt.text = text
        root_annotate.append(annotation_txt)

        return root_annotate

    def _create_relation(self, relation_type, annotations):
        root_relation = etree.Element('relation')
        root_relation.set("id", "R"+str(self._relations_id))

        infon = etree.Element('infon')
        infon.set("key", "type")
        infon.text = relation_type
        root_relation.append(infon)

        infon = etree.Element('infon')
        infon.set("key", "annotator")
        infon.text = "DODFMiner"
        root_relation.append(infon)

        infon = etree.Element('infon')
        infon.set("key", "updated_at")
        infon.text = str(datetime.datetime.now())
        root_relation.append(infon)

        for ann_id in annotations:
            node = etree.Element('node')
            node.set("refid", str(ann_id))
            node.set("role", "")
            root_relation.append(node)

        return root_relation



"""NER backend for act and propriety extraction.

This module contains the ActNER class, which have all that is necessary to
extract an act and, its proprieties, using a trained ner model.

"""

import numpy as np
import os
import joblib
from nltk import word_tokenize

# pylint: disable=too-few-public-methods
class ActNER:
    """Act NER Class.

    This class encapsulate all functions, and attributes related
    to the process of NER extraction.

    Note:
        This class is one of the fathers of the Base act class.

    Attributes:
        _model: The trained NER model for the act

    """

    def __init__(self):
        # self._backend = 'regex'
        super().__init__()


        # pylint: disable=assignment-from-no-return
        self._model = self._load_model()

    def _load_model(self):
        """Load Model from models/folder.

        Note:
            This function needs to be overwriten in
            the child class. If this function is not
            overwrite the backend will fall back to regex.

        """
        # pylint: disable=access-member-before-definition
        if self._backend == 'ner':
            print(f"Act {self._name} does not have a model: FALLING BACK TO REGEX")
            self._backend = 'regex'
        else:
            self._backend = 'regex'

    @classmethod
    def number_of_digits(self, s):
        """ Returns number of digits in string. """
        return sum(c.isdigit() for c in s)

    @classmethod
    def get_base_feat(self, word):
        """ Returns base features of a word. """
        d = {
            'word': word.lower(),
            'is_title': word.istitle(),
            'is_upper': word.isupper(),
            'num_digits': str(self.number_of_digits(word)),
        }
        return d

    @classmethod
    def _get_features(cls, sentence):
        """Create features for each word in act.

        Create a list of dict of words features to be used in the predictor module.

        Args:
            act (list): List of words in an act.

        Returns:
            A list with a dictionary of features for each of the words.

        """
        sent_features = []

        for i in range(len(sentence)):
            word = sentence[i]
    
            
            word_before = '' if i == 0 else sentence[i-1]
            word_before2 = '' if i <= 1 else sentence[i-2]
            word_before3 = '' if i <= 2 else sentence[i-3]
            
            word_after = '' if i+1 == len(sentence) else sentence[i+1]
            word_after2 = '' if i+2 >= len(sentence) else sentence[i+2]
            word_after3 = '' if i+3 >= len(sentence) else sentence[i+3]
        
            word_before = cls.get_base_feat(word_before)
            word_before2 = cls.get_base_feat(word_before2)
            word_before3 = cls.get_base_feat(word_before3)
            word_after = cls.get_base_feat(word_after)
            word_after2 = cls.get_base_feat(word_after2)
            word_after3 = cls.get_base_feat(word_after3)
            
            word_feat = {
                'bias': 1.0,
                'word': word.lower(),
                'is_title': word.istitle(),
                'is_upper': word.isupper(),
                'is_digit': word.isdigit(),
                
                'num_digits': str(cls.number_of_digits(word)),
                'has_hyphen': '-' in word,
                'has_dot': '.' in word,
                'has_slash': '/' in word,
            }
            
            if i > 0:
                word_feat.update({
                    '-1:word': word_before['word'].lower(),
                    '-1:title': word_before['is_title'],
                    '-1:upper': word_before['is_upper'],
                    '-1:num_digits': word_before['num_digits'],
                })
            else:
                word_feat['BOS'] = True
                
            if i > 1:
                word_feat.update({
                    '-2:word': word_before2['word'].lower(),
                    '-2:title': word_before2['is_title'],
                    '-2:upper': word_before2['is_upper'],
                    '-2:num_digits': word_before2['num_digits'],
                })
                
            if i > 2:
                word_feat.update({
                    '-3:word': word_before3['word'].lower(),
                    '-3:title': word_before3['is_title'],
                    '-3:upper': word_before3['is_upper'],
                    '-3:num_digits': word_before3['num_digits'],
                })
                
            if i < len(sentence) - 1:
                word_feat.update({
                    '+1:word': word_after['word'].lower(),
                    '+1:title': word_after['is_title'],
                    '+1:upper': word_after['is_upper'],
                    '+1:num_digits': word_after['num_digits'],
                })
            else:
                word_feat['EOS'] = True
                
            if i < len(sentence) - 2:
                word_feat.update({
                    '+2:word': word_after2['word'].lower(),
                    '+2:title': word_after2['is_title'],
                    '+2:upper': word_after2['is_upper'],
                    '+2:num_digits': word_after2['num_digits'],
                })
                
            if i < len(sentence) - 3:
                word_feat.update({
                    '+3:word': word_after3['word'].lower(),
                    '+3:title': word_after3['is_title'],
                    '+3:upper': word_after3['is_upper'],
                    '+3:num_digits': word_after3['num_digits'],
                })
                
            sent_features.append(word_feat)
        
        return sent_features

    def add_common_attributes(self, feats):
        f_path = os.path.dirname(os.path.dirname(__file__))
        f_path += '/acts/models/atributos_gerais.pkl'
        crf = joblib.load(f_path)

        predictions = crf.predict_single(feats)

        for i in range(len(feats)):
            feats[i]['predicao'] = predictions[i]

    def _prediction(self, act):
        """Predict classes for a single act.

        Args:
            act (string): Full act

        Returns:
            A dictionary with the proprieties and its
            predicted value.
        """
        print("Predicting")
        act = self._preprocess(act)

        feats = self._get_features(act)
        self.add_common_attributes(feats)

        predictions = self._model.predict_single(feats)
        return self._predictions_dict(act, predictions)

    @classmethod
    def _preprocess(cls, sentence):
        """Transform a raw string to a list of words.

        Args:
            sentence (string): Raw act as string.

        Returns:
            The list of words in the act.
        """ 
        sentence = word_tokenize(sentence.replace('/', ' / ').replace(':', ' : ').replace('``', ' ').replace("''", ' '))
        return sentence

    def _predictions_dict(self, sentence, prediction):
        """Create dictionary of proprieties.

        Create dictionary of tags to save predicted entities.

        Args:
            sentence (list): List of words and tokens in the act.
            prediction ([type]): The correspondent predicitons for each
                                 word in the sentence.

        Returns:
            A dictionary of the proprieties found.

        """
        tags = self._model.classes_
        tags.remove('O')
        tags.sort(reverse=True)
        while tags[0][0] == 'I':
            del tags[0]
        for i, _ in enumerate(tags):
            tags[i] = tags[i][2:]
        dict_ato = {}
        dict_ato["Tipo do Ato"] = ""
        for i in tags:
            dict_ato[i] = []

        last_tag = 'O'
        temp_entity = []
        for i, _ in enumerate(prediction):
            if last_tag not in (prediction[i], 'O'):
                dict_ato[last_tag[2:]].append(temp_entity)
                temp_entity = []
            if prediction[i] == 'O':
                last_tag = 'O'
                continue
            # else:
            temp_entity.append(sentence[i])
            last_tag = "I" + prediction[i][1:]

        if temp_entity:
            dict_ato[last_tag[2:]].append(temp_entity)

        for key, ato in dict_ato.items():
            values = []
            for value in ato:
                value = ' '.join(value)
                values.append(value)
            if len(values) >= 1:
                dict_ato[key] = values[0]


            if dict_ato[key] == []:
                dict_ato[key] = np.nan

        # pylint: disable=no-member
        dict_ato["Tipo do Ato"] = self._name


        return dict_ato



# pylint: disable=protected-access

import os
from unittest.mock import patch
import pytest
import numpy as np
import pandas as pd
from dodfminer.extract.polished.acts.base import Atos

rule_dict = {
    "numeros": r"([0-9]+)",
    "capitalizado": r"([A-Z][a-z]+)",
    "nao_encontra": "khalil"
}
RULE_PROP = r"(MENSAGEM:)([^.]*)"
props = ["Tipo", "Numeros", "Capitalizado", "Não Encontra"]
valid_file = ""+os.path.dirname(__file__)+"/support/valid.txt"

TEXT = "fghngfnfgnfgnfgnrr MENSAGEM: O Renato testa esse codigo 1232 vezes por dia. qeqcnoecqucqpwxqrthrweqqeprto MENSAGEM: " + \
    "O Joao testa esse codigo 2 vezes por dia. wsfhsn wenrwermrkne MENSAGEM: O Lucas testa esse codigo 0 vezes por dia. efwefwefwfwefwetrynyujuju"
act = [' O Renato testa esse codigo 1232 vezes por dia',
       ' O Joao testa esse codigo 2 vezes por dia', ' O Lucas testa esse codigo 0 vezes por dia'
       ]

res_act = [
    {
        'tipo_ato': 'Aposentadoria',
        'numeros': '1232',
        'capitalizado': 'Renato',
        'nao_encontra': np.nan
    }, {
        'tipo_ato': 'Aposentadoria',
        'numeros': '2',
        'capitalizado': 'Joao',
        'nao_encontra': np.nan
    }, {
        'tipo_ato': 'Aposentadoria',
        'numeros': '0',
        'capitalizado': 'Lucas',
        'nao_encontra': np.nan
    }
]


# Wrong Init Type
@patch.object(Atos, '_act_name', return_value="Aposentadoria")
@patch.object(Atos, '_props_names', return_value=props)
@patch.object(Atos, '_rule_for_inst', return_value=RULE_PROP)
@patch.object(Atos, '_prop_rules', return_value=rule_dict)
def test_act_base_invalid_file(*_):
    file = ""+os.path.dirname(__file__)+"/support/invalid.pdf"
    atos = Atos(file, 'regex')
    assert atos._file_name is None
    assert atos._text == file


@patch.object(Atos, '_act_name', return_value="Aposentadoria")
@patch.object(Atos, '_props_names', return_value=props)
@patch.object(Atos, '_rule_for_inst', return_value=RULE_PROP)
@patch.object(Atos, '_prop_rules', return_value=rule_dict)
def test_act_non_existing_backend(*_):
    ato = Atos(valid_file, 'lol')
    assert ato._backend == 'regex'


@patch.object(Atos, '_act_name', return_value="Aposentadoria")
@patch.object(Atos, '_props_names', return_value=props)
@patch.object(Atos, '_rule_for_inst', return_value=RULE_PROP)
@patch.object(Atos, '_prop_rules', return_value=rule_dict)
def test_act_backend_fallback(*_):
    ato = Atos(valid_file, 'ner')
    assert ato._backend == 'regex'

# Wrong Init


@patch.object(Atos, '_act_name', return_value="Aposentadoria")
@patch.object(Atos, '_props_names', return_value=props)
@patch.object(Atos, '_rule_for_inst', return_value=RULE_PROP)
def test_no_overwrite_1(*_):
    with pytest.raises(Exception):
        Atos(valid_file, 'regex')


@patch.object(Atos, '_act_name', return_value="Aposentadoria")
@patch.object(Atos, '_props_names', return_value=props)
@patch.object(Atos, '_prop_rules', return_value=rule_dict)
def test_no_overwrite_2(*_):
    with pytest.raises(Exception):
        Atos(valid_file, 'regex')


@patch.object(Atos, '_act_name', return_value="Aposentadoria")
@patch.object(Atos, '_rule_for_inst', return_value=RULE_PROP)
@patch.object(Atos, '_prop_rules', return_value=rule_dict)
def test_no_overwrite_3(*_):
    with pytest.raises(Exception):
        Atos(valid_file, 'regex')


@patch.object(Atos, '_props_names', return_value=props)
@patch.object(Atos, '_rule_for_inst', return_value=RULE_PROP)
@patch.object(Atos, '_prop_rules', return_value=rule_dict)
def test_no_overwrite_4(*_):
    with pytest.raises(Exception):
        Atos(valid_file, 'regex')


def test_no_overwrite_5():
    with pytest.raises(Exception):
        Atos(valid_file, 'regex')


@pytest.fixture(name='act_base_regex')
@patch.object(Atos, '_act_name', return_value="Aposentadoria")
@patch.object(Atos, '_props_names', return_value=["Numeros", "Capitalizado", "Não Encontra"])
@patch.object(Atos, '_rule_for_inst',
              return_value=r"(MENSAGEM:)([^.]*)")
@patch.object(Atos, '_prop_rules', return_value={
    "numeros": r"([0-9]+)",
    "capitalizado": r"([A-Z][a-z]+)",
    "nao_encontra": "khalil"
})
def fixture_act_base_regex(*_):
    atos = Atos(valid_file, 'regex')
    atos._name = "Aposentadoria"
    return atos


def test_act_base_name(act_base_regex):
    assert act_base_regex.name == "Aposentadoria"


def test_act_base_name_attr(act_base_regex):
    assert act_base_regex._name == "Aposentadoria"


def test_act_file_name(act_base_regex):
    assert act_base_regex._file_name == valid_file


def test_act_base_raw_acts_regex(act_base_regex):
    act_base_regex._backend = 'regex'
    act_base_regex._text = TEXT
    assert act_base_regex._regex_instances() == act


def test_act_base_raw_acts_ner(act_base_regex):
    act_base_regex._backend = 'ner'
    act_base_regex._text = TEXT
    assert act_base_regex._regex_instances() == act


def test_act_base_raw_acts_error(act_base_regex):
    act_base_regex._backend = 'lol'
    act_base_regex._text = TEXT
    with pytest.raises(Exception):
        act_base_regex._extract_instances()


def test_act_base_props_regex(act_base_regex):
    act_base_regex._backend = 'regex'
    act_base_regex._raw_acts = act
    print(act_base_regex._acts)
    assert act_base_regex._extract_props() == res_act


def test_act_base_build_dataframes(act_base_regex):
    act_base_regex._backend = 'regex'
    act_base_regex._acts = res_act
    act_base_regex._columns = props
    res_df = pd.DataFrame(res_act)
    res_df.columns = props

    assert act_base_regex._build_dataframe().equals(res_df)


def test_act_base_build_dataframes_empty(act_base_regex):
    act_base_regex._backend = 'regex'
    res_df = pd.DataFrame()
    assert act_base_regex._build_dataframe().equals(res_df)


def test_act_base_props_error(act_base_regex):
    act_base_regex._backend = 'lol'
    act_base_regex._raw_acts = act
    with pytest.raises(Exception):
        act_base_regex._extract_props()



import os
import json
import sys
from glob import glob
import shutil
from unittest.mock import patch
import pytest
from dodfminer.run import run

EMPTY_ARGS_MSG = """usage: DODFMiner [-h] {downloader,extract} ...

Data extractor of PDF documents from the Official Gazette of the Federal
District, Brazil.

positional arguments:
  {downloader,extract}

optional arguments:
  -h, --help            show this help message and exit

© Copyright 2020, KnEDLe Team. Version 1.1.3.4
"""

ONLY_EXTRACTED_MSG = "usage: DODFMiner extract"

act_choices = ["aposentadoria",
               "nomeacao",
               "exoneracao",
               "abono",
               "retificacoes",
               "reversoes",
               "substituicao",
               "efetivos_nome",
               "efetivos_exo"]


def test_run_no_args():
    with pytest.raises(SystemExit):
        with patch.object(sys, 'argv', ['cmd', '']):
            run()


def test_run_only_extract_arg(capsys):
    targets = ['cmd', 'extract']
    with patch.object(sys, 'argv', targets):
        run()
        captured = capsys.readouterr()
        assert ONLY_EXTRACTED_MSG in captured.out


def test_run_extract_single_file_pure_text():
    file = ""+os.path.dirname(__file__)+"/support/dodfminer_sf.pdf"
    targets = ["cmd", "extract", "-s", file, "-t", "pure-text"]
    content_txt = "NOMEAR JOSE ANTONIO BARBOSA, Tecnico de Gestao Educacional,"
    with patch.object(sys, 'argv', targets):
        run()
        assert os.path.exists(file.replace('pdf', 'txt'))
        with open(file.replace('pdf', 'txt'), encoding='utf-8') as file_txt:
            assert content_txt in file_txt.read()
        os.remove(file.replace('pdf', 'txt'))


def test_run_extract_single_file_with_titles():
    file = "" + os.path.dirname(__file__) + "/support/dodfminer_sf.pdf"
    targets = ["cmd", "extract", "-s", file, "-t", "with-titles"]

    with patch.object(sys, 'argv', targets):
        run()

        assert os.path.exists(file.replace('pdf', 'json'))

        with open(file.replace('pdf', 'json'), encoding='utf-8') as json_pointer:
            json_file = json_pointer.read()
            res_dict = json.loads(json_file)

        assert "SECAO I" in res_dict.keys()
        assert "PODER EXECUTIVO" in res_dict['SECAO I'].keys()

        os.remove(file.replace('pdf', 'json'))

def test_run_extract_single_file_blocks():
    file = ""+os.path.dirname(__file__)+"/support/dodfminer_sf.pdf"
    targets = ["cmd", "extract", "-s", file, "-t", "blocks"]
    with patch.object(sys, 'argv', targets):
        run()
        assert os.path.exists(file.replace('pdf', 'json'))
        with open(file.replace('pdf', 'json'), encoding='utf-8') as json_file:
            res_dict = json.loads(json_file.read())
            assert len(res_dict) > 0
        os.remove(file.replace('pdf', 'json'))


def test_run_extract_single_all_act():
    folder = ""+os.path.dirname(__file__)+"/support/"
    file = folder+"dodfminer_sf.pdf"
    targets = ["cmd", "extract", "-s", file, "-a"]
    with patch.object(sys, 'argv', targets):
        run()
        for act in act_choices:
            assert os.path.isfile(folder+act+".csv")
            os.remove(folder+act+".csv")


def test_run_extract_input_folder_pure_text():
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs"
    targets = ["cmd", "extract", "-i", folder, "-t", "pure-text"]
    with patch.object(sys, 'argv', targets):
        run()
        res_folder = folder + '/results/txt/2020/01_Janeiro/'
        assert os.path.isdir(res_folder)
        assert len(glob(res_folder+'*.txt')) > 1
        shutil.rmtree(folder + '/results/')


def test_run_extract_input_folder_with_titles():
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs"
    targets = ["cmd", "extract", "-i", folder, "-t", "with-titles"]
    with patch.object(sys, 'argv', targets):
        run()
        res_folder = folder + '/results/json/2020/01_Janeiro/'
        assert os.path.isdir(res_folder)
        assert len(glob(res_folder+'*.json')) > 1
        shutil.rmtree(folder + '/results/')


def test_run_extract_input_folder_one_act():
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs"
    targets = ["cmd", "extract", "-i", folder, "-a", "aposentadoria"]
    with patch.object(sys, 'argv', targets):
        run()
        assert os.path.isfile(folder+"/aposentadoria.csv")
        os.remove(folder+"/aposentadoria.csv")


def test_run_extract_input_folder_two_act():
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs"
    targets = ["cmd", "extract", "-i", folder, "-a", "aposentadoria", "abono"]

    with patch.object(sys, 'argv', targets):
        run()

        assert os.path.isfile(folder+"/aposentadoria.csv")
        assert os.path.isfile(folder+"/abono.csv")

        os.remove(folder+"/aposentadoria.csv")
        os.remove(folder+"/abono.csv")


def test_run_extract_input_folder_all_act():
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs/"
    targets = ["cmd", "extract", "-i", folder, "-a"]
    with patch.object(sys, 'argv', targets):
        run()
        for act in act_choices:
            assert os.path.isfile(folder+act+".csv")
            os.remove(folder+act+".csv")
    shutil.rmtree(folder + '/results')


def test_run_extract_input_folder_one_act_back_end_ner():
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs"
    targets = ["cmd", "extract", "-i", folder,
               "-a", "aposentadoria", "-b", "ner"]
    with patch.object(sys, 'argv', targets):
        run()
        assert os.path.isfile(folder+"/aposentadoria.csv")
        os.remove(folder+"/aposentadoria.csv")


def test_run_extract_single_file_committee_classification():
    folder = ""+os.path.dirname(__file__)+"/support/"
    file = folder+"dodfminer_sf.pdf"
    targets = ["cmd", "extract", "-s", file, "-a", "-c"]
    with patch.object(sys, 'argv', targets):
        run()
        for act in act_choices:
            assert os.path.isfile(folder+act+".csv")
            os.remove(folder+act+".csv")

def test_run_extract_input_folder_committee_classification():
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs/"
    targets = ["cmd", "extract", "-i", folder, "-a", "-c"]
    with patch.object(sys, 'argv', targets):
        run()
        for act in act_choices:
            assert os.path.isfile(folder+act+".csv")
            os.remove(folder+act+".csv")
    shutil.rmtree(folder + '/results')

def test_run_extract_input_folder_xml():
    folder = ""+os.path.dirname(__file__)+"/support/xml_extract"
    targets = ["cmd", "extract", "-i", folder, "-x"]
    with patch.object(sys, 'argv', targets):
        run()
        assert os.path.isfile(os.path.join(folder, '1_10.1.2020.xml'))
        os.remove(folder+'/1_10.1.2020.xml')

# def test_run_extract_input_single_xml():
#   file = ""+os.path.dirname(__file__)+"/support/xml_extract/DODF 001 02-01-2020 INTEGRA.pdf"
#   targets = ["cmd", "extract", "-s", file, "-x"]
#   with patch.object(sys, 'argv', targets):
#     run()
#     res_file = ""+os.path.dirname(__file__)+"/support/xml_extract/1_1.2.1.2020.xml"
#     assert os.path.isfile(res_file)
#     os.remove(res_file)



# pylint: disable=protected-access

import os
from unittest.mock import patch
import joblib
import pytest
from dodfminer.extract.polished.backend.seg import ActSeg

# act_seg with regex segmentation

@pytest.fixture(name='act_seg_regex')
@patch.object(ActSeg, '_backend', create = True)
def fixture_act_seg_regex(*_):
    act = ActSeg()
    act._name = 'test'
    act._backend = 'regex'

    act._flags = 0
    act._rules = {
        "numeros": r"([0-9]+)",
        "capitalizado": r"([A-Z][a-z]+)",
        "nao_encontra": "khalil"
    }
    act._inst_rule = r"(MENSAGEM:)([^.]*)"
    act._acts_str = []

    act.__init__()
    return act

def test_act_seg_regex(act_seg_regex):
    assert act_seg_regex._backend == 'regex'
    assert act_seg_regex._seg_function == act_seg_regex._regex_instances

# act_seg with ner segmentation

@pytest.fixture(name='act_seg_ner')
@patch.object(ActSeg, '_backend', create = True)
@patch.object(ActSeg, '_load_seg_model',
              return_value=joblib.load(""+os.path.dirname(__file__)+"/support/test_seg_model.pkl"))
def fixture_act_seg_ner(*_):
    act = ActSeg()
    act._name = 'test'
    act._backend = 'ner'
    act.__init__()
    return act

def test_act_seg_ner(act_seg_ner):
    assert act_seg_ner._backend == 'ner'
    assert act_seg_ner._seg_function == act_seg_ner._crf_instances

# act_seg with ner background but no model, reverting to regex segmentation

@pytest.fixture(name='act_seg_reverting_to_regex')
@patch.object(ActSeg, '_backend', create = True)
def fixture_act_seg_reverting_to_regex(*_):
    act = ActSeg()
    act._name = 'test'
    act._backend = 'ner'
    act.__init__()
    return act

def test_act_seg_reverting_to_regex(act_seg_reverting_to_regex):
    assert act_seg_reverting_to_regex._backend == 'ner'
    assert act_seg_reverting_to_regex._seg_function == act_seg_reverting_to_regex._regex_instances

#

def test_act_seg_regex_instances(act_seg_regex):
    act_seg_regex._text =  "fghngfnfgnfgnfgnrr MENSAGEM: O Renato testa esse codigo 1232 vezes por dia. qeqcnoecqucqpwxqrthrweqqeprto MENSAGEM: " + \
        "O Joao testa esse codigo 2 vezes por dia. wsfhsn wenrwermrkne MENSAGEM: O Lucas testa esse codigo 0 vezes por dia. efwefwefwfwefwetrynyujuju"
    act_seg_regex.acts_str = []
    act = act_seg_regex._seg_function()

    assert act == [' O Renato testa esse codigo 1232 vezes por dia',
                   ' O Joao testa esse codigo 2 vezes por dia', ' O Lucas testa esse codigo 0 vezes por dia']

def test_act_seg_crf_instances(act_seg_ner):
    act_seg_ner._text = "CONCEDER, aposentadoria voluntária integral, ao servidor ELIAS SANTOS MONTEIRO, matrícula nº 24.679-4, no cargo de " + \
        "Analista em Políticas Públicas e Gestão Governamental, Classe Especial, Padrão V, do Quadro de Pessoal do Distrito Federal, nos termos " + \
        "do artigo 3º, incisos I, II e III, e Parágrafo único da Emenda Constitucional nº 47 de 05/07/2005, combinado com o artigo 44 da Lei " + \
        "Complementar nº 769, de 30/06/2008, e com a vantagem pessoal prevista no artigo 5º da Lei nº 4.584, de 08/07/2011. Lotação: " + \
        "Administração Regional de Brazlândia. Processo SEI nº 00133-00002749/2020-19. CONCEDER, aposentadoria voluntária integral, à servidora " + \
        "MIRIAM RODRIGUES DA SILVA, matrícula nº 31.325-4, no cargo de Analista em Políticas Públicas e Gestão Governamental, Classe Especial, " + \
        "Padrão V, do Quadro de Pessoal do Distrito Federal, nos termos do artigo 3º, incisos I, II e III, e Parágrafo único da Emenda " + \
        "Constitucional nº 47 de 05/07/2005, combinado com o artigo 44 da Lei Complementar nº 769, de 30/06/2008, e com a vantagem pessoal " + \
        "prevista no artigo 5º da Lei nº 4.584, de 08/07/2011. Lotação: Administração Regional de Brazlândia. Processo SEI nº " + \
        "00133-00002735/2020-97"
    act_seg_ner._acts_str = []
    act = act_seg_ner._seg_function()

    assert len(act) == 2

def test_act_seg_preprocess(act_seg_ner):
    text = "CONCEDER, aposentadoria voluntária integral, ao servidor ELIAS SANTOS MONTEIRO, matrícula nº 24.679-4, no cargo de " + \
        "Analista em Políticas Públicas e Gestão Governamental, Classe Especial, Padrão V, do Quadro de Pessoal do Distrito Federal, nos termos " + \
        "do artigo 3º, incisos I, II e III, e Parágrafo único da Emenda Constitucional nº 47 de 05/07/2005, combinado com o artigo 44 da Lei " + \
        "Complementar nº 769, de 30/06/2008, e com a vantagem pessoal prevista no artigo 5º da Lei nº 4.584, de 08/07/2011. Lotação: " + \
        "Administração Regional de Brazlândia. Processo SEI nº 00133-00002749/2020-19."
    preprocessed_text = ['CONCEDER', ',', 'aposentadoria', 'voluntária', 'integral', ',', 'ao', 'servidor', 'ELIAS', 'SANTOS', 'MONTEIRO', ',',
        'matrícula', 'nº', '24', '.', '679', '-', '4', ',', 'no', 'cargo', 'de', 'Analista', 'em', 'Políticas', 'Públicas', 'e', 'Gestão',
        'Governamental', ',', 'Classe', 'Especial', ',', 'Padrão', 'V', ',', 'do', 'Quadro', 'de', 'Pessoal', 'do', 'Distrito', 'Federal', ',',
        'nos', 'termos', 'do', 'artigo', '3º', ',', 'incisos', 'I', ',', 'II', 'e', 'III', ',', 'e', 'Parágrafo', 'único', 'da', 'Emenda',
        'Constitucional', 'nº', '47', 'de', '05', '/', '07', '/', '2005', ',', 'combinado', 'com', 'o', 'artigo', '44', 'da', 'Lei', 'Complementar',
        'nº', '769', ',', 'de', '30', '/', '06', '/', '2008', ',', 'e', 'com', 'a', 'vantagem', 'pessoal', 'prevista', 'no', 'artigo', '5º', 'da',
        'Lei', 'nº', '4', '.', '584', ',', 'de', '08', '/', '07', '/', '2011', '.', 'Lotação', ':', 'Administração', 'Regional', 'de', 'Brazlândia',
        '.', 'Processo', 'SEI', 'nº', '00133', '-', '00002749', '/', '2020', '-', '19', '.']
    assert act_seg_ner._preprocess(text) == preprocessed_text

def test_act_seg_number_of_digits(act_seg_ner):
    text = "123abc456"
    assert act_seg_ner._number_of_digits(text) == 6


def test_act_seg_get_base_feat(act_seg_ner):
    word = "NER"
    assert act_seg_ner._get_base_feat(word) == {
        'word': 'ner',
        'is_title': False,
        'is_upper': True,
        'num_digits': '0',
    }


def test_act_seg_get_features(act_seg_ner):
    text = "CONCEDER , servidor ELIAS 24.679-4".split()
    features = [
        {
            'bias': 1.0, 'word': 'conceder', 'is_title': False, 'is_upper': True, 'num_digits': '0', 'BOS': True, '+1:word': ',', '+1:title': False,
            '+1:upper': False, '+1:num_digits': '0', '+2:word': 'servidor', '+2:title': False, '+2:upper': False, '+2:num_digits': '0',
            '+3:word': 'elias', '+3:title': False, '+3:upper': True, '+3:num_digits': '0', '+4:word': '24.679-4', '+4:title': False,
            '+4:upper': False, '+4:num_digits': '6'
        },
        {
            'bias': 1.0, 'word': ',', 'is_title': False, 'is_upper': False, 'num_digits': '0', '-1:word': 'conceder', '-1:title': False,
            '-1:upper': True, '-1:num_digits': '0', '+1:word': 'servidor', '+1:title': False, '+1:upper': False, '+1:num_digits': '0',
            '+2:word': 'elias', '+2:title': False, '+2:upper': True, '+2:num_digits': '0', '+3:word': '24.679-4', '+3:title': False,
            '+3:upper': False, '+3:num_digits': '6'
        },
        {
            'bias': 1.0, 'word': 'servidor', 'is_title': False, 'is_upper': False, 'num_digits': '0', '-1:word': ',', '-1:title': False,
            '-1:upper': False, '-1:num_digits': '0', '-2:word': 'conceder', '-2:title': False, '-2:upper': True, '-2:num_digits': '0',
            '+1:word': 'elias', '+1:title': False, '+1:upper': True, '+1:num_digits': '0', '+2:word': '24.679-4', '+2:title': False,
            '+2:upper': False, '+2:num_digits': '6'
        },
        {
            'bias': 1.0, 'word': 'elias', 'is_title': False, 'is_upper': True, 'num_digits': '0', '-1:word': 'servidor', '-1:title': False,
            '-1:upper': False, '-1:num_digits': '0', '-2:word': ',', '-2:title': False, '-2:upper': False, '-2:num_digits': '0',
            '-3:word': 'conceder', '-3:title': False, '-3:upper': True, '-3:num_digits': '0', '+1:word': '24.679-4', '+1:title': False,
            '+1:upper': False, '+1:num_digits': '6'
        },
        {
            'bias': 1.0, 'word': '24.679-4', 'is_title': False, 'is_upper': False, 'num_digits': '6', '-1:word': 'elias', '-1:title': False,
            '-1:upper': True, '-1:num_digits': '0', '-2:word': 'servidor', '-2:title': False, '-2:upper': False, '-2:num_digits': '0',
            '-3:word': ',', '-3:title': False, '-3:upper': False, '-3:num_digits': '0', '-4:word': 'conceder', '-4:title': False,
            '-4:upper': True, '-4:num_digits': '0', 'EOS': True
        }]

    assert act_seg_ner._get_features(text) == features


def test_act_seg_extract_acts(act_seg_ner):
    text = "AAAA  BBBB  CCCC  DDDD EEEE FFFF  XXXX  YYYY  ZZZZ".split()
    pred = "B-Ato I-Ato E-ato O    O    B-Ato I-Ato E-Ato O".split()
    alvo =["AAAA BBBB CCCC",            "FFFF XXXX YYYY"]

    assert act_seg_ner._extract_acts(text, pred) == alvo



"""Base class for act segmentation.

This module contains the ActSeg class, which have all that is necessary to
extract acts from a block of text.
"""

import re
from nltk import word_tokenize

class ActSeg:
    """Base class for act segmentation.

    This class encapsulate all functions, and attributes related
    to the process of segmentation of an act.

    Note:
        This class is one of the fathers of the Base act class.

    Attributes:
        _seg_function: Function for segmentation.

    """

    def __init__(self):
        self._seg_function = self._load_seg_function()

    def _load_seg_function(self):
        """Load segmentation function into the _seg_function variable.

        If the segmentation backend is ner and a NER model is available, loads the function _crf_instances.
        Otherwise, loads the function _regex_instances.

        """
        if self._backend == 'ner':
            self._seg_model = self._load_seg_model()
            if self._seg_model is not None:
                return self._crf_instances
            else:
                print(f"Act {self._name} does not have a segmentation model: Using regex for segmentation")
                return self._regex_instances
        else:
            self._backend = 'regex'
            return self._regex_instances
    
    def _load_seg_model(self):
        """Load Segmentation Model from models/folder.

        Note:
            This function needs to be overwriten in
            the child class. If this function is not
            overwrite the segmentation backend will fall back to regex.

        """
        return None

    def _regex_instances(self):
        """Search for all instances of the act using the defined rule.

        Returns:
            List of all act instances in the text.
        """

        # pylint: disable=no-member
        found = re.findall(self._inst_rule, self._text, flags=self._flags)
        results = []
        for instance in found:
            head, body, *_ = instance
            # pylint: disable=no-member
            self._acts_str.append(head+body)
            results.append(body)

        return results

    def _crf_instances(self):
        """Search for all instances of the act using a CRF model.
        
        Returns:
            List of all act instances in the text.
        """
        text = self._preprocess(self._text)
        feats = self._get_features(text)
        pred = self._seg_model.predict_single(feats)
        acts = self._extract_acts(text, pred)
        self._acts_str += acts
        return acts

    def _preprocess(self, text):
        """Preprocess text for CRF model."""
        return word_tokenize(text.replace('-', ' - ').replace('/', ' / ').replace('.', ' . ').replace(',', ' , '))

    def _number_of_digits(self, s):
        """Returns number of digits in a string."""
        return sum(c.isdigit() for c in s)

    def _get_base_feat(self, word):
        """Returns basic features for a word, used in the CRF model."""
        d = {
            'word': word.lower(),
            'is_title': word.istitle(),
            'is_upper': word.isupper(),
            'num_digits': str(self._number_of_digits(word)),
        }
        return d

    def _get_features(self, text):
        """Create features for each word in a text.

        Args:
            text (list): List of words in an act.
        Returns:
            A list with a dictionary of features for each of the words.
        """
        features = []
            
        for i in range(len(text)):
            word = text[i]
            
            word_before = '' if i == 0 else text[i-1]
            word_before2 = '' if i <= 1 else text[i-2]
            word_before3 = '' if i <= 2 else text[i-3]
            word_before4 = '' if i <= 3 else text[i-4]
            
            word_after = '' if i+1 == len(text) else text[i+1]
            word_after2 = '' if i+2 >= len(text) else text[i+2]
            word_after3 = '' if i+3 >= len(text) else text[i+3]
            word_after4 = '' if i+4 >= len(text) else text[i+4]
            
            word_before = self._get_base_feat(word_before)
            word_before2 = self._get_base_feat(word_before2)
            word_before3 = self._get_base_feat(word_before3)
            word_before4 = self._get_base_feat(word_before4)
            word_after = self._get_base_feat(word_after)
            word_after2 = self._get_base_feat(word_after2)
            word_after3 = self._get_base_feat(word_after3)
            word_after4 = self._get_base_feat(word_after4)
            
            word_feat = {
                'bias': 1.0,
                'word': word.lower(),
                'is_title': word.istitle(),
                'is_upper': word.isupper(),
                'num_digits': str(self._number_of_digits(word)),
            }
            
            if i > 0:
                word_feat.update({
                    '-1:word': word_before['word'].lower(),
                    '-1:title': word_before['is_title'],
                    '-1:upper': word_before['is_upper'],
                    '-1:num_digits': word_before['num_digits'],
                })
            else:
                word_feat['BOS'] = True
                
            if i > 1:
                word_feat.update({
                    '-2:word': word_before2['word'].lower(),
                    '-2:title': word_before2['is_title'],
                    '-2:upper': word_before2['is_upper'],
                    '-2:num_digits': word_before2['num_digits'],
                })
                
            if i > 2:
                word_feat.update({
                    '-3:word': word_before3['word'].lower(),
                    '-3:title': word_before3['is_title'],
                    '-3:upper': word_before3['is_upper'],
                    '-3:num_digits': word_before3['num_digits'],
                })

            if i > 3:
                word_feat.update({
                    '-4:word': word_before4['word'].lower(),
                    '-4:title': word_before4['is_title'],
                    '-4:upper': word_before4['is_upper'],
                    '-4:num_digits': word_before4['num_digits'],
                })
                
            if i < len(text) - 1:
                word_feat.update({
                    '+1:word': word_after['word'].lower(),
                    '+1:title': word_after['is_title'],
                    '+1:upper': word_after['is_upper'],
                    '+1:num_digits': word_after['num_digits'],
                })
            else:
                word_feat['EOS'] = True
                
            if i < len(text) - 2:
                word_feat.update({
                    '+2:word': word_after2['word'].lower(),
                    '+2:title': word_after2['is_title'],
                    '+2:upper': word_after2['is_upper'],
                    '+2:num_digits': word_after2['num_digits'],
                })
                
            if i < len(text) - 3:
                word_feat.update({
                    '+3:word': word_after3['word'].lower(),
                    '+3:title': word_after3['is_title'],
                    '+3:upper': word_after3['is_upper'],
                    '+3:num_digits': word_after3['num_digits'],
                })

            if i < len(text) - 4:
                word_feat.update({
                    '+4:word': word_after4['word'].lower(),
                    '+4:title': word_after4['is_title'],
                    '+4:upper': word_after4['is_upper'],
                    '+4:num_digits': word_after4['num_digits'],
                })
                
            features.append(word_feat)
        
        return features

    def _extract_acts(self, text, prediction):
        """Extract and join words predicted to be part of an act.
        
        Args:
            text (list): List of words in an act.
            prediction (list): Predictions made for each word in the act.
            
        """
        acts = []
        
        current_act = []
        reading_act = False
        for i in range(len(prediction)):
            if prediction[i][0] == 'B': #B-ato
                reading_act = True
                current_act.append(text[i])
                continue
                
            if reading_act:
                current_act.append(text[i])
                
            if reading_act and prediction[i][0] == 'E': #E-ato
                acts.append(' '.join(current_act))
                current_act = []
                reading_act = False
                
        if reading_act:
            acts.append(' '.join(current_act))
            
        return acts


"""Base class for an Act model.

This module contains the Atos class, which have all that is necessary to
extract information from a specialized act.
"""

import pandas as pd

from dodfminer.extract.polished.backend.regex import ActRegex
from dodfminer.extract.polished.backend.ner import ActNER
from dodfminer.extract.polished.backend.seg import ActSeg


class Atos(ActRegex, ActNER, ActSeg):
    """Base class for extracting an act and its proprieties to a dataframe.

    Note:
        You should not use this class alone,
        use its childs on the regex module.

    Args:
        file (str): The dodf file path.
        backend (str): The mechanism to use in extraction.
                       Can be either regex or ner.
                       Defaults to regex.

    Attributes:
        _file_name (str): The dodf file path.
        _text (str): The dodf content in string format.
        _acts_str (str): List of raw text acts.
        _name (str): Name of the act.
        _columns (str): List of the proprieties names from the act.
        _raw_acts (list): List of raw text acts .
        _acts (list): List of acts with propreties extracted.
        _data_frame (dataframe): The resulting dataframe from the
                                 extraction process.

    """

    def __init__(self, file, backend='regex'):
        self._backend = backend
        self._name = self._act_name()
        super().__init__()

        try:
            with open(file, "r", encoding='utf-8') as fp:
                self._text = fp.read()
                fp.close()
                self._file_name = file
        except IOError:
            self._text = file
            self._file_name = None

        self._acts_str = []
        self._columns = self._props_names()

        self._raw_acts = self._seg_function()
        self._acts = self._extract_props()
        self._data_frame = self._build_dataframe()

    @property
    def name(self):
        """str: Name of the act."""
        return self._name

    @property
    def data_frame(self):
        """:obj:`dataframe`: Act dataframe with proprieties extracted."""
        return self._data_frame

    @property
    def acts_str(self):
        """str: Vector of acts content as raw text."""
        return self._acts_str

    def _act_name(self):
        """Name of the act.

        Must return a single string representing the act name

        Raises:
            NotImplementedError: Child class needs to overwrite this method.

        """
        raise NotImplementedError

    def _props_names(self):
        """Name of all the proprieties for the dataframe column.

        Must return a vector of string representing the proprieties names

        Warning:
            The first name will be used for the type-of-act propriety.

        Raises:
            NotImplementedError: Child class needs to overwrite this method.

        """
        raise NotImplementedError

    def _build_dataframe(self):
        """Create a dataframe with the extracted proprieties.

        Returns:
            The dataframe created
        """
        if len(self._acts) > 0:
            data_frame = pd.DataFrame(self._acts)
            if self._backend == 'regex':
                data_frame.columns = self._columns
            else:
                data_frame.columns = [x.capitalize() for x in data_frame.columns]
            return data_frame
        return pd.DataFrame()

    def _extract_props(self):
        """Extract proprieties of all the acts.

        Returns:
            A vector of extracted acts dictionaries.
        """
        acts = []
        for value in self._raw_acts:
            act = {}
            if self._backend == 'regex':
                act = self._regex_props(value)
            elif self._backend == 'ner':
                act = self._prediction(value)
            else:
                raise NotImplementedError("Non-existent backend option")
            acts.append(act)
        return acts


import os

from dodfminer.extract.polished.core import ActsExtractor, _acts_ids

from dodfminer.extract.polished.acts.aposentadoria import Retirements, RetAposentadoria
from dodfminer.extract.polished.acts.cessoes import Cessoes
from dodfminer.extract.polished.acts.nomeacao import NomeacaoComissionados, NomeacaoEfetivos
from dodfminer.extract.polished.acts.exoneracao import Exoneracao, ExoneracaoEfetivos
from dodfminer.extract.polished.acts.reversoes import Revertions
from dodfminer.extract.polished.acts.abono import AbonoPermanencia
from dodfminer.extract.polished.acts.sem_efeito_aposentadoria import SemEfeitoAposentadoria
from dodfminer.extract.polished.acts.substituicao import Substituicao
from dodfminer.extract.polished.acts.contrato import Contratos


from dodfminer.extract.polished.create_xml import XMLFy


def test_polished_core_acts():
    assert _acts_ids == {
        "aposentadoria": Retirements,
        "reversoes": Revertions,
        "nomeacao": NomeacaoComissionados,
        "exoneracao": Exoneracao,
        "abono": AbonoPermanencia,
        "retificacoes": RetAposentadoria,
        "substituicao": Substituicao,
        "efetivos_nome": NomeacaoEfetivos,
        "efetivos_exo": ExoneracaoEfetivos,
        "sem_efeito_aposentadoria": SemEfeitoAposentadoria,
        "cessoes": Cessoes,
        "contrato": Contratos
    }


def test_polished_core_act_obj():
    assert isinstance(ActsExtractor.get_act_obj(
        "aposentadoria", "", "regex"), Retirements)
    assert isinstance(ActsExtractor.get_act_obj(
        "reversoes", "", "regex"), Revertions)
    assert isinstance(ActsExtractor.get_act_obj(
        "nomeacao", "", "regex"), NomeacaoComissionados)
    assert isinstance(ActsExtractor.get_act_obj(
        "abono", "", "regex"), AbonoPermanencia)
    assert isinstance(ActsExtractor.get_act_obj(
        "retificacoes", "", "regex"), RetAposentadoria)
    assert isinstance(ActsExtractor.get_act_obj(
        "substituicao", "", "regex"), Substituicao)
    assert isinstance(ActsExtractor.get_act_obj(
        "efetivos_nome", "", "regex"), NomeacaoEfetivos)
    assert isinstance(ActsExtractor.get_act_obj(
        "efetivos_exo", "", "regex"), ExoneracaoEfetivos)
    assert isinstance(ActsExtractor.get_act_obj(
        "exoneracao", "", "regex"), Exoneracao)
    assert isinstance(ActsExtractor.get_act_obj(
        "sem_efeito_aposentadoria", "", "regex"), SemEfeitoAposentadoria)
    assert isinstance(ActsExtractor.get_act_obj(
        "cessoes", "", "regex"), Cessoes)
    assert isinstance(ActsExtractor.get_act_obj(
        "contrato", "", "regex"), Contratos)
    assert len(_acts_ids) == 12


def test_polished_core_get_all_obj():
    objs = ActsExtractor.get_all_obj("", "regex")
    needed = {
        "abono": AbonoPermanencia("", "regex"),
        "aposentadoria": Retirements("", "regex"),
        "efetivos_exo": ExoneracaoEfetivos("", "regex"),
        "efetivos_nome": NomeacaoEfetivos("", "regex"),
        "exoneracao": Exoneracao("", "regex"),
        "nomeacao": NomeacaoComissionados("", "regex"),
        "retificacoes": RetAposentadoria("", "regex"),
        "reversoes": Revertions("", "regex"),
        "substituicao": Substituicao("", "regex"),
        "sem_efeito_aposentadoria": SemEfeitoAposentadoria("", "regex"),
        "cessoes": Cessoes("", "regex"),
        "contrato": Contratos("", "regex")
    }
    assert len(objs) == 12
    assert len(objs) == len(needed)


def test_polished_core_act_df():
    assert isinstance(ActsExtractor.get_act_df("aposentadoria", "", "regex"), type(Retirements("", "regex").data_frame))
    assert isinstance(ActsExtractor.get_act_df("reversoes", "", "regex"), type(Revertions("", "regex").data_frame))
    assert isinstance(ActsExtractor.get_act_df("nomeacao", "", "regex"), type(NomeacaoComissionados("", "regex").data_frame))
    assert isinstance(ActsExtractor.get_act_df("abono", "", "regex"), type(AbonoPermanencia("", "regex").data_frame))
    assert isinstance(ActsExtractor.get_act_df("retificacoes", "", "regex"), type(RetAposentadoria("", "regex").data_frame))
    assert isinstance(ActsExtractor.get_act_df("substituicao", "", "regex"), type(Substituicao("", "regex").data_frame))
    assert isinstance(ActsExtractor.get_act_df("efetivos_nome", "", "regex"), type(NomeacaoEfetivos("", "regex").data_frame))
    assert isinstance(ActsExtractor.get_act_df("efetivos_exo", "", "regex"), type(ExoneracaoEfetivos("", "regex").data_frame))
    assert isinstance(ActsExtractor.get_act_df("exoneracao", "", "regex"), type(Exoneracao("", "regex").data_frame))
    assert isinstance(ActsExtractor.get_act_df("sem_efeito_aposentadoria", "", "regex"), type(SemEfeitoAposentadoria("", "regex").data_frame))
    assert isinstance(ActsExtractor.get_act_df("cessoes", "", "regex"), type(Cessoes("", "regex").data_frame))
    assert isinstance(ActsExtractor.get_act_df("contrato", "", "regex"), type(Contratos("", "regex").data_frame))
    assert len(_acts_ids) == 12


def test_polished_core_get_all_df():
    data_frames = ActsExtractor.get_all_df("", "regex")
    dataframes = {
        "abono": AbonoPermanencia("", "regex").data_frame,
        "aposentadoria": Retirements("", "regex").data_frame,
        "efetivos_exo": ExoneracaoEfetivos("", "regex").data_frame,
        "efetivos_nome": NomeacaoEfetivos("", "regex").data_frame,
        "exoneracao": Exoneracao("", "regex").data_frame,
        "nomeacao": NomeacaoComissionados("", "regex").data_frame,
        "retificacoes": RetAposentadoria("", "regex").data_frame,
        "reversoes": Revertions("", "regex").data_frame,
        "substituicao": Substituicao("", "regex").data_frame,
        "sem_efeito_aposentadoria": SemEfeitoAposentadoria("", "regex").data_frame,
        "cessoes": Cessoes("", "regex").data_frame,
        "contrato": Contratos("", "regex").data_frame
    }
    assert len(data_frames) == 12
    assert len(data_frames) == len(dataframes)


def test_polished_get_xml():
    xml = ActsExtractor.get_xml(""+os.path.dirname(__file__) +
                                "/support/DODF 001 01-01-2019 EDICAO ESPECIAL.pdf", "regex", 0)
    assert isinstance(xml, XMLFy)



import os
import json
import shutil

from glob import glob
from dodfminer.extract.pure.core import ContentExtractor

EXPECTED_EXTRACTED_TEXT = "BRASILIA - DF, QUINTA-FEIRA, 2 DE JANEIRO DE 2020 SUMARIO\nSECAO I"
DODF_FILE_PATH = file = "" + \
    os.path.dirname(__file__)+"/support/dodfminer_sf.pdf"


def test_pure_extract_text_single():
    txt_file_path = DODF_FILE_PATH.replace("pdf", "txt")

    ContentExtractor.extract_text(DODF_FILE_PATH, single=True)

    assert os.path.isfile(txt_file_path)
    with open(txt_file_path, encoding='utf-8') as txt_file:
        assert EXPECTED_EXTRACTED_TEXT in txt_file.read()

    os.remove(txt_file_path)


def test_pure_extract_text_return_text():
    assert EXPECTED_EXTRACTED_TEXT in ContentExtractor.extract_text(
        DODF_FILE_PATH)


def test_pure_extract_text_return_list():
    file_pdf = ""+os.path.dirname(__file__)+"/support/dodfminer_sf.pdf"
    # generated_txt = "BRASILIA - DF, QUINTA-FEIRA, 2 DE JANEIRO DE 2020 SECAO I SUMARIO"
    assert len(ContentExtractor.extract_text(file_pdf, block=True)) > 0


def test_pure_extract_text_single_return_list():
    json_file_path = DODF_FILE_PATH.replace("pdf", "json")

    ContentExtractor.extract_text(DODF_FILE_PATH, single=True, block=True)

    assert os.path.isfile(json_file_path)
    with open(json_file_path, encoding='utf-8') as json_file:
        assert len(json.loads(json_file.read())) > 0

    os.remove(json_file_path)


def test_pure_extract_text_json_false_saves_txt_file():
    txt_file_path = DODF_FILE_PATH.replace("pdf", "txt")

    ContentExtractor.extract_text(
        DODF_FILE_PATH, single=True, block=True, is_json=False)

    assert os.path.isfile(txt_file_path)
    with open(txt_file_path, encoding='utf-8') as txt_file:
        content = txt_file.read()
        assert isinstance(content, str)
        assert len(content) > 0

    os.remove(txt_file_path)


# def test_pure_extract_text_norm():
#     file = ""+os.path.dirname(__file__)+"/support/dodfminer_sf.pdf"
#     generated_txt = "BRASILIA - DF, QUINTA-FEIRA, 2 DE JANEIRO DE 2020 SECAO I SUMARIO"
#     assert unicodedata.is_normalized("NFC", ContentExtractor.extract_text(file, norm="NFC"))


def test_pure_extract_structure_single():
    pdf_file = ""+os.path.dirname(__file__)+"/support/dodfminer_sf.pdf"
    json_file = pdf_file.replace("pdf", "json")

    ContentExtractor.extract_structure(pdf_file, single=True)

    assert os.path.isfile(json_file)

    with open(json_file, encoding='utf-8') as file_pointer:
        file_content = file_pointer.read()
        assert 'SECAO I' in json.loads(file_content).keys()
        assert 'PODER EXECUTIVO' in json.loads(file_content)['SECAO I'].keys()

    os.remove(file.replace("pdf", "json"))

def test_pure_extract_structure():
    pdf_file = ""+os.path.dirname(__file__)+"/support/dodfminer_sf.pdf"

    assert 'SECAO I' in ContentExtractor.extract_structure(pdf_file)
    assert 'PODER EXECUTIVO' in ContentExtractor.extract_structure(pdf_file)['SECAO I']


def test_pure_extract_to_txt():
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs"
    res_folder = folder + '/results/txt/2020/01_Janeiro/'
    ContentExtractor.extract_to_txt(folder)
    assert os.path.isdir(res_folder)
    assert len(glob(res_folder+'*.txt')) > 1
    shutil.rmtree(folder + '/results/')


def test_pure_extract_to_json_without_titles():
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs"
    res_folder = folder + '/results/json/2020/01_Janeiro/'
    ContentExtractor.extract_to_json(folder)
    assert os.path.isdir(res_folder)
    assert len(glob(res_folder+'*.json')) > 1
    shutil.rmtree(folder + '/results/')


def test_pure_extract_to_json_with_titles():
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs"
    res_folder = folder + '/results/json/2020/01_Janeiro/'
    ContentExtractor.extract_to_json(folder, titles_with_boxes=True)
    assert os.path.isdir(res_folder)
    assert len(glob(res_folder+'*.json')) > 1
    shutil.rmtree(folder + '/results/')


def test_pure_extract_to_json_already_exists(capsys):
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs"
    # res_folder = folder + '/results/json/2020/01_Janeiro/'
    ContentExtractor.extract_to_json(folder, titles_with_boxes=True)
    ContentExtractor.extract_to_json(folder, titles_with_boxes=True)
    captured = capsys.readouterr()
    assert "JSON already exists" in captured.out
    shutil.rmtree(folder + '/results/')


def test_pure_extract_to_txt_already_exists(capsys):
    folder = ""+os.path.dirname(__file__)+"/support/dodf_pdfs"
    # res_folder = folder + '/results/txt/2020/01_Janeiro/'
    ContentExtractor.extract_to_txt(folder)
    ContentExtractor.extract_to_txt(folder)
    captured = capsys.readouterr()
    assert "TXT already exists" in captured.out
    shutil.rmtree(folder + '/results/')
